"""Communicate with an Android TV device via ADB over a network.

ADB Debugging must be enabled.
"""


import logging

from ..basetv.basetv import BaseTV
from .. import constants

_LOGGER = logging.getLogger(__name__)


class BaseAndroidTV(BaseTV):  # pylint: disable=too-few-public-methods
    """Representation of an Android TV device.

    Parameters
    ----------
    host : str
        The address of the device; may be an IP address or a host name
    port : int
        The device port to which we are connecting (default is 5555)
    adbkey : str
        The path to the ``adbkey`` file for ADB authentication
    adb_server_ip : str
        The IP address of the ADB server
    adb_server_port : int
        The port for the ADB server
    state_detection_rules : dict, None
        A dictionary of rules for determining the state (see :class:`~androidtv.basetv.basetv.BaseTV`)

    """

    DEVICE_CLASS = "androidtv"
    DEVICE_ENUM = constants.DeviceEnum.ANDROIDTV

    def __init__(self, host, port=5555, adbkey="", adb_server_ip="", adb_server_port=5037, state_detection_rules=None):
        BaseTV.__init__(self, None, host, port, adbkey, adb_server_ip, adb_server_port, state_detection_rules)

    # ======================================================================= #
    #                                                                         #
    #                          Home Assistant Update                          #
    #                                                                         #
    # ======================================================================= #
    def _update(
        self,
        screen_on,
        awake,
        audio_state,
        wake_lock_size,
        current_app,
        media_session_state,
        audio_output_device,
        is_volume_muted,
        volume,
        running_apps,
        hdmi_input,
    ):
        """Get the info needed for a Home Assistant update.

        Parameters
        ----------
        screen_on : bool, None
            Whether or not the device is on, or ``None`` if it was not determined
        awake : bool, None
            Whether or not the device is awake (screensaver is not running), or ``None`` if it was not determined
        audio_state : str, None
            The audio state, as determined from "dumpsys audio", or ``None`` if it was not determined
        wake_lock_size : int, None
            The size of the current wake lock, or ``None`` if it was not determined
        current_app : str, None
            The current app property, or ``None`` if it was not determined
        media_session_state : int, None
            The state from the output of ``dumpsys media_session``, or ``None`` if it was not determined
        audio_output_device : str, None
            The current audio playback device, or ``None`` if it was not determined
        is_volume_muted : bool, None
            Whether or not the volume is muted, or ``None`` if it was not determined
        volume : int, None
            The absolute volume level, or ``None`` if it was not determined
        running_apps : list, None
            A list of the running apps, or ``None`` if it was not determined
        hdmi_input : str, None
            The HDMI input, or ``None`` if it could not be determined

        Returns
        -------
        state : str
            The state of the device
        current_app : str
            The current running app
        running_apps : list
            A list of the running apps if ``get_running_apps`` is True, otherwise the list ``[current_app]``
        audio_output_device : str
            The current audio playback device
        is_volume_muted : bool
            Whether or not the volume is muted
        volume_level : float
            The volume level (between 0 and 1)
        hdmi_input : str, None
            The HDMI input, or ``None`` if it could not be determined

        """
        # Get the volume (between 0 and 1)
        volume_level = self._volume_level(volume)

        # Check if device is unavailable
        if screen_on is None:
            state = None

        # Check if device is off
        elif not screen_on or current_app == "off":
            state = constants.STATE_OFF

        # Check if screen saver is on
        elif not awake:
            state = constants.STATE_STANDBY

        else:
            # Get the running apps
            if not running_apps and current_app:
                running_apps = [current_app]

            # Determine the state using custom rules
            state = self._custom_state_detection(
                current_app=current_app,
                media_session_state=media_session_state,
                wake_lock_size=wake_lock_size,
                audio_state=audio_state,
            )
            if state:
                return state, current_app, running_apps, audio_output_device, is_volume_muted, volume_level, hdmi_input

            # ATV Launcher
            if current_app in [constants.APP_ATV_LAUNCHER, None]:
                state = constants.STATE_IDLE

            # BELL Fibe
            elif current_app == constants.APP_BELL_FIBE:
                state = audio_state

            # Netflix
            elif current_app == constants.APP_NETFLIX:
                if media_session_state == 2:
                    state = constants.STATE_PAUSED
                elif media_session_state == 3:
                    state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_IDLE

            # NLZIET
            elif current_app == constants.APP_NLZIET:
                if wake_lock_size == 1:
                    state = constants.STATE_PAUSED
                elif wake_lock_size == 2:
                    state = constants.STATE_PLAYING
            # Plex
            elif current_app == constants.APP_PLEX:
                if media_session_state == 3:
                    if wake_lock_size == 1:
                        state = constants.STATE_PAUSED
                    else:
                        state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_IDLE

            # TVheadend
            elif current_app == constants.APP_TVHEADEND:
                if wake_lock_size == 5:
                    state = constants.STATE_PAUSED
                elif wake_lock_size == 6:
                    state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_IDLE

            # VLC
            elif current_app == constants.APP_VLC:
                if media_session_state == 2:
                    state = constants.STATE_PAUSED
                elif media_session_state == 3:
                    state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_IDLE

            # VRV
            elif current_app == constants.APP_VRV:
                state = audio_state

            # YouTube
            elif current_app == constants.APP_YOUTUBE:
                if media_session_state == 2:
                    state = constants.STATE_PAUSED
                elif media_session_state == 3:
                    state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_IDLE

            # Get the state from `media_session_state`
            elif media_session_state:
                if media_session_state == 2:
                    state = constants.STATE_PAUSED
                elif media_session_state == 3:
                    state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_IDLE

            # Get the state from `audio_state`
            elif audio_state != constants.STATE_IDLE:
                state = audio_state

            # Get the state from `wake_lock_size`
            else:
                if wake_lock_size == 1:
                    state = constants.STATE_PAUSED
                elif wake_lock_size == 2:
                    state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_IDLE

        return state, current_app, running_apps, audio_output_device, is_volume_muted, volume_level, hdmi_input



"""Communicate with an Android TV or Amazon Fire TV device via ADB over a network.

ADB Debugging must be enabled.
"""


import logging
import re

from .. import constants

_LOGGER = logging.getLogger(__name__)


class BaseTV(object):  # pylint: disable=too-few-public-methods
    """Base class for representing an Android TV / Fire TV device.

    The ``state_detection_rules`` parameter is of the format:

    .. code-block:: python

       state_detection_rules = {'com.amazon.tv.launcher': ['idle'],
                                'com.netflix.ninja': ['media_session_state'],
                                'com.ellation.vrv': ['audio_state'],
                                'com.hulu.plus': [{'playing': {'wake_lock_size' : 4}},
                                                  {'paused': {'wake_lock_size': 2}}],
                                'com.plexapp.android': [{'paused': {'media_session_state': 3, 'wake_lock_size': 1}},
                                                        {'playing': {'media_session_state': 3}},
                                                        'idle']}

    The keys are app IDs, and the values are lists of rules that are evaluated in order.

    :py:const:`~androidtv.constants.VALID_STATES`

    .. code-block:: python

       VALID_STATES = ('idle', 'off', 'playing', 'paused', 'standby')


    **Valid rules:**

    * ``'idle'``, ``'playing'``, ``'paused'``, ``'standby'``, or ``'off'`` = always report the specified state when this app is open
    * ``'media_session_state'`` = try to use the :meth:`media_session_state` property to determine the state
    * ``'audio_state'`` = try to use the :meth:`audio_state` property to determine the state
    * ``{'<VALID_STATE>': {'<PROPERTY1>': VALUE1, '<PROPERTY2>': VALUE2, ...}}`` = check if each of the properties is equal to the specified value, and if so return the state

      * The valid properties are ``'media_session_state'``, ``'audio_state'``, and ``'wake_lock_size'``


    Parameters
    ----------
    adb : ADBPythonSync, ADBServerSync, ADBPythonAsync, ADBServerAsync
        The handler for ADB commands
    host : str
        The address of the device; may be an IP address or a host name
    port : int
        The device port to which we are connecting (default is 5555)
    adbkey : str
        The path to the ``adbkey`` file for ADB authentication
    adb_server_ip : str
        The IP address of the ADB server
    adb_server_port : int
        The port for the ADB server
    state_detection_rules : dict, None
        A dictionary of rules for determining the state (see above)

    """

    DEVICE_ENUM = constants.DeviceEnum.BASETV

    def __init__(
        self,
        adb,
        host,
        port=5555,
        adbkey="",
        adb_server_ip="",
        adb_server_port=5037,
        state_detection_rules=None,
    ):
        self._adb = adb
        self.host = host
        self.port = int(port)
        self.adbkey = adbkey
        self.adb_server_ip = adb_server_ip
        self.adb_server_port = adb_server_port
        self._state_detection_rules = state_detection_rules
        self.device_properties = {}
        self.installed_apps = []

        # make sure the rules are valid
        if self._state_detection_rules:
            for app_id, rules in self._state_detection_rules.items():
                if not isinstance(app_id, str):
                    raise TypeError("{0} is of type {1}, not str".format(app_id, type(app_id).__name__))
                state_detection_rules_validator(rules)

        # the max volume level (determined when first getting the volume level)
        self.max_volume = None

        # Customizable commands
        self._custom_commands = {}

    # ======================================================================= #
    #                                                                         #
    #                      Device-specific ADB commands                       #
    #                                                                         #
    # ======================================================================= #
    def customize_command(self, custom_command, value):
        """Customize a command used to retrieve properties.

        Parameters
        ----------
        custom_command : str
            The name of the command that will be customized; it must be in `constants.CUSTOMIZABLE_COMMANDS`
        value : str, None
            The custom ADB command that will be used, or ``None`` if the custom command should be deleted

        """
        if custom_command in constants.CUSTOMIZABLE_COMMANDS:
            if value is not None:
                self._custom_commands[custom_command] = value
            elif custom_command in self._custom_commands:
                del self._custom_commands[custom_command]

    def _cmd_audio_state(self):
        """Get the command used to retrieve the current audio state for this device.

        Returns
        -------
        str
            The device-specific ADB shell command used to determine the current audio state

        """
        if constants.CUSTOM_AUDIO_STATE in self._custom_commands:
            return self._custom_commands[constants.CUSTOM_AUDIO_STATE]

        # Is this an Android 11 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "11":
            return constants.CMD_AUDIO_STATE11

        # Is this an Android 12 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "12":
            return constants.CMD_AUDIO_STATE11

        # Is this an Android 13 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "13":
            return constants.CMD_AUDIO_STATE11
        return constants.CMD_AUDIO_STATE

    def _cmd_current_app(self):
        """Get the command used to retrieve the current app for this device.

        Returns
        -------
        str
            The device-specific ADB shell command used to determine the current app

        """
        if constants.CUSTOM_CURRENT_APP in self._custom_commands:
            return self._custom_commands[constants.CUSTOM_CURRENT_APP]

        # Is this a Google Chromecast Android TV?
        if (
            self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV
            and "Google" in self.device_properties.get("manufacturer", "")
            and "Chromecast" in self.device_properties.get("model", "")
        ):
            return constants.CMD_CURRENT_APP_GOOGLE_TV

        # Is this an Android 11 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "11":
            return constants.CMD_CURRENT_APP11

        # Is this an Android 12 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "12":
            return constants.CMD_CURRENT_APP12

        # Is this an Android 13 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "13":
            return constants.CMD_CURRENT_APP13

        return constants.CMD_CURRENT_APP

    def _cmd_current_app_media_session_state(self):
        """Get the command used to retrieve the current app and media session state for this device.

        Returns
        -------
        str
            The device-specific ADB shell command used to determine the current app and media session state

        """
        if constants.CUSTOM_CURRENT_APP_MEDIA_SESSION_STATE in self._custom_commands:
            return self._custom_commands[constants.CUSTOM_CURRENT_APP_MEDIA_SESSION_STATE]

        # Is this a Google Chromecast Android TV?
        if (
            self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV
            and "Google" in self.device_properties.get("manufacturer", "")
            and "Chromecast" in self.device_properties.get("model", "")
        ):
            return constants.CMD_CURRENT_APP_MEDIA_SESSION_STATE_GOOGLE_TV

        # Is this an Android 11 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "11":
            return constants.CMD_CURRENT_APP_MEDIA_SESSION_STATE11

        # Is this an Android 12 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "12":
            return constants.CMD_CURRENT_APP_MEDIA_SESSION_STATE12

        # Is this an Android 13 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "13":
            return constants.CMD_CURRENT_APP_MEDIA_SESSION_STATE13

        return constants.CMD_CURRENT_APP_MEDIA_SESSION_STATE

    def _cmd_hdmi_input(self):
        """Get the command used to retrieve the current HDMI input for this device.

        Returns
        -------
        str
            The device-specific ADB shell command used to determine the current HDMI input

        """
        if constants.CUSTOM_HDMI_INPUT in self._custom_commands:
            return self._custom_commands[constants.CUSTOM_HDMI_INPUT]

        # Is this an Android 11 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "11":
            return constants.CMD_HDMI_INPUT11

        # Is this an Android 12 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "12":
            return constants.CMD_HDMI_INPUT11

        # Is this an Android 13 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "13":
            return constants.CMD_HDMI_INPUT11

        return constants.CMD_HDMI_INPUT

    def _cmd_volume_set(self, new_volume):
        """Get the command used to set volume for this device.

        Parameters
        ----------
        new_volume : int
            The new volume level

        Returns
        -------
        str
            The device-specific ADB shell command used to set volume

        """
        # Is this an Android 11 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "11":
            return constants.CMD_VOLUME_SET_COMMAND11.format(new_volume)

        # Is this an Android 12 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "12":
            return constants.CMD_VOLUME_SET_COMMAND11.format(new_volume)

        # Is this an Android 13 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "13":
            return constants.CMD_VOLUME_SET_COMMAND11.format(new_volume)

        return constants.CMD_VOLUME_SET_COMMAND.format(new_volume)

    def _cmd_launch_app(self, app):
        """Get the command to launch the specified app for this device.

        Parameters
        ----------
        app : str
            The app that will be launched

        Returns
        -------
        str
            The device-specific command to launch the app

        """
        if constants.CUSTOM_LAUNCH_APP in self._custom_commands:
            return self._custom_commands[constants.CUSTOM_LAUNCH_APP].format(app)

        # Is this a Google Chromecast Android TV?
        if (
            self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV
            and "Google" in self.device_properties.get("manufacturer", "")
            and "Chromecast" in self.device_properties.get("model", "")
        ):
            return constants.CMD_LAUNCH_APP_GOOGLE_TV.format(app)

        if self.DEVICE_ENUM == constants.DeviceEnum.FIRETV:
            return constants.CMD_LAUNCH_APP_FIRETV.format(app)

        # Is this an Android 11 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "11":
            return constants.CMD_LAUNCH_APP11.format(app)

        # Is this an Android 12 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "12":
            return constants.CMD_LAUNCH_APP12.format(app)

        # Is this an Android 13 device?
        if self.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV and self.device_properties.get("sw_version", "") == "13":
            return constants.CMD_LAUNCH_APP13.format(app)

        return constants.CMD_LAUNCH_APP.format(app)

    def _cmd_running_apps(self):
        """Get the command used to retrieve the running apps for this device.

        Returns
        -------
        str
            The device-specific ADB shell command used to determine the running apps

        """
        if constants.CUSTOM_RUNNING_APPS in self._custom_commands:
            return self._custom_commands[constants.CUSTOM_RUNNING_APPS]

        if self.DEVICE_ENUM == constants.DeviceEnum.FIRETV:
            return constants.CMD_RUNNING_APPS_FIRETV

        return constants.CMD_RUNNING_APPS_ANDROIDTV

    def _cmd_turn_off(self):
        """Get the command used to turn off this device.

        Returns
        -------
        str
            The device-specific ADB shell command used to turn off the device

        """
        if constants.CUSTOM_TURN_OFF in self._custom_commands:
            return self._custom_commands[constants.CUSTOM_TURN_OFF]

        if self.DEVICE_ENUM == constants.DeviceEnum.FIRETV:
            return constants.CMD_TURN_OFF_FIRETV

        return constants.CMD_TURN_OFF_ANDROIDTV

    def _cmd_turn_on(self):
        """Get the command used to turn on this device.

        Returns
        -------
        str
            The device-specific ADB shell command used to turn on the device

        """
        if constants.CUSTOM_TURN_ON in self._custom_commands:
            return self._custom_commands[constants.CUSTOM_TURN_ON]

        if self.DEVICE_ENUM == constants.DeviceEnum.FIRETV:
            return constants.CMD_TURN_ON_FIRETV

        return constants.CMD_TURN_ON_ANDROIDTV

    # ======================================================================= #
    #                                                                         #
    #                               ADB methods                               #
    #                                                                         #
    # ======================================================================= #
    @property
    def available(self):
        """Whether the ADB connection is intact.

        Returns
        -------
        bool
            Whether or not the ADB connection is intact

        """
        return self._adb.available

    @staticmethod
    def _remove_adb_shell_prefix(cmd):
        """Remove the 'adb shell ' prefix from ``cmd``, if present.

        Parameters
        ----------
        cmd : str
            The ADB shell command

        Returns
        -------
        str
            ``cmd`` with the 'adb shell ' prefix removed, if it was present

        """
        return cmd[len("adb shell ") :] if cmd.startswith("adb shell ") else cmd

    # ======================================================================= #
    #                                                                         #
    #                        Home Assistant device info                       #
    #                                                                         #
    # ======================================================================= #
    def _parse_device_properties(self, properties):
        """Return a dictionary of device properties.

        Parameters
        ----------
        properties : str, None
            The output of the ADB command that retrieves the device properties

        This method fills in the ``device_properties`` attribute, which is a dictionary with keys
        ``'serialno'``, ``'manufacturer'``, ``'model'``, and ``'sw_version'``

        """
        _LOGGER.debug(
            "%s:%d `get_device_properties` response: %s",
            self.host,
            self.port,
            properties,
        )

        if not properties:
            self.device_properties = {}
            return

        lines = properties.strip().splitlines()
        if len(lines) != 4:
            self.device_properties = {}
            return

        manufacturer, model, serialno, version = lines

        if not serialno.strip():
            _LOGGER.warning(
                "Could not obtain serialno for %s:%d, got: '%s'",
                self.host,
                self.port,
                serialno,
            )
            serialno = None

        self.device_properties = {
            "manufacturer": manufacturer,
            "model": model,
            "serialno": serialno,
            "sw_version": version,
        }

    @staticmethod
    def _parse_mac_address(mac_response):
        """Parse a MAC address from the ADB shell response.

        Parameters
        ----------
        mac_response : str, None
            The response from the MAC address ADB shell command

        Returns
        -------
        str, None
            The parsed MAC address, or ``None`` if it could not be determined

        """
        if not mac_response:
            return None

        mac_matches = re.findall(constants.MAC_REGEX_PATTERN, mac_response)
        if mac_matches:
            return mac_matches[0]

        return None

    # ======================================================================= #
    #                                                                         #
    #                         Custom state detection                          #
    #                                                                         #
    # ======================================================================= #
    def _custom_state_detection(
        self,
        current_app=None,
        media_session_state=None,
        wake_lock_size=None,
        audio_state=None,
    ):
        """Use the rules in ``self._state_detection_rules`` to determine the state.

        Parameters
        ----------
        current_app : str, None
            The :meth:`current_app` property
        media_session_state : int, None
            The :meth:`media_session_state` property
        wake_lock_size : int, None
            The :meth:`wake_lock_size` property
        audio_state : str, None
            The :meth:`audio_state` property

        Returns
        -------
        str, None
            The state, if it could be determined using the rules in ``self._state_detection_rules``; otherwise, ``None``

        """
        if not self._state_detection_rules or current_app is None or current_app not in self._state_detection_rules:
            return None

        rules = self._state_detection_rules[current_app]

        for rule in rules:
            # The state is always the same for this app
            if rule in constants.VALID_STATES:
                return rule

            # Use the `media_session_state` property
            if rule == "media_session_state":
                if media_session_state == 2:
                    return constants.STATE_PAUSED
                if media_session_state == 3:
                    return constants.STATE_PLAYING
                if media_session_state is not None:
                    return constants.STATE_IDLE

            # Use the `audio_state` property
            if rule == "audio_state" and audio_state in constants.VALID_STATES:
                return audio_state

            # Check conditions and if they are true, return the specified state
            if isinstance(rule, dict):
                for state, conditions in rule.items():
                    if state in constants.VALID_STATES and self._conditions_are_true(
                        conditions, media_session_state, wake_lock_size, audio_state
                    ):
                        return state

        return None

    @staticmethod
    def _conditions_are_true(conditions, media_session_state=None, wake_lock_size=None, audio_state=None):
        """Check whether the conditions in ``conditions`` are true.

        Parameters
        ----------
        conditions : dict
            A dictionary of conditions to be checked (see the ``state_detection_rules`` parameter in :class:`~androidtv.basetv.basetv.BaseTV`)
        media_session_state : int, None
            The :meth:`media_session_state` property
        wake_lock_size : int, None
            The :meth:`wake_lock_size` property
        audio_state : str, None
            The :meth:`audio_state` property

        Returns
        -------
        bool
            Whether or not all the conditions in ``conditions`` are true

        """
        for key, val in conditions.items():
            if key == "media_session_state":
                if media_session_state is None or media_session_state != val:
                    return False

            elif key == "wake_lock_size":
                if wake_lock_size is None or wake_lock_size != val:
                    return False

            elif key == "audio_state":
                if audio_state is None or audio_state != val:
                    return False

            # key is invalid
            else:
                return False

        return True

    # ======================================================================= #
    #                                                                         #
    #                            Parse properties                             #
    #                                                                         #
    # ======================================================================= #
    @staticmethod
    def _audio_output_device(stream_music):
        """Get the current audio playback device from the ``STREAM_MUSIC`` block from ``adb shell dumpsys audio``.

        Parameters
        ----------
        stream_music : str, None
            The ``STREAM_MUSIC`` block from ``adb shell dumpsys audio``

        Returns
        -------
        str, None
            The current audio playback device, or ``None`` if it could not be determined

        """
        if not stream_music:
            return None

        matches = re.findall(constants.DEVICE_REGEX_PATTERN, stream_music, re.DOTALL | re.MULTILINE)
        if matches:
            return matches[0]

        return None

    @staticmethod
    def _audio_state(audio_state_response):
        """Parse the :meth:`audio_state` property from the ADB shell output.

        Parameters
        ----------
        audio_state_response : str, None
            The output from the ADB command `androidtv.basetv.basetv.BaseTV._cmd_audio_state``

        Returns
        -------
        str, None
            The audio state, or ``None`` if it could not be determined

        """
        if not audio_state_response:
            return None
        if audio_state_response == "1":
            return constants.STATE_PAUSED
        if audio_state_response == "2":
            return constants.STATE_PLAYING
        return constants.STATE_IDLE

    @staticmethod
    def _current_app(current_app_response):
        """Get the current app from the output of the command `androidtv.basetv.basetv.BaseTV._cmd_current_app`.

        Parameters
        ----------
        current_app_response : str, None
            The output from the ADB command `androidtv.basetv.basetv.BaseTV._cmd_current_app`

        Returns
        -------
        str, None
            The current app, or ``None`` if it could not be determined

        """
        if not current_app_response or "=" in current_app_response or "{" in current_app_response:
            return None

        return current_app_response

    def _current_app_media_session_state(self, current_app_media_session_state_response):
        """Get the current app and the media session state properties from the output of `androidtv.basetv.basetv.BaseTV._cmd_current_app_media_session_state`.

        Parameters
        ----------
        current_app_media_session_state_response : str, None
            The output of `androidtv.basetv.basetv.BaseTV._cmd_current_app_media_session_state`

        Returns
        -------
        current_app : str, None
            The current app, or ``None`` if it could not be determined
        media_session_state : int, None
            The state from the output of the ADB shell command, or ``None`` if it could not be determined

        """
        if not current_app_media_session_state_response:
            return None, None

        lines = current_app_media_session_state_response.splitlines()

        current_app = self._current_app(lines[0].strip())

        if len(lines) > 1:
            matches = constants.REGEX_MEDIA_SESSION_STATE.search(current_app_media_session_state_response)
            if matches:
                return current_app, int(matches.group("state"))

        return current_app, None

    @staticmethod
    def _get_hdmi_input(hdmi_response):
        """Get the HDMI input from the from the ADB shell output`.

        Parameters
        ----------
        hdmi_response : str, None
            The output from the ADB command `androidtv.basetv.basetv.BaseTV._cmd_hdmi_input``

        Returns
        -------
        str, None
            The HDMI input, or ``None`` if it could not be determined

        """
        return hdmi_response.strip() if hdmi_response and hdmi_response.strip() else None

    @staticmethod
    def _get_installed_apps(installed_apps_response):
        """Get the installed apps from the output of :py:const:`androidtv.constants.CMD_INSTALLED_APPS`.

        Parameters
        ----------
        installed_apps_response : str, None
            The output of :py:const:`androidtv.constants.CMD_INSTALLED_APPS`

        Returns
        -------
        list, None
            A list of the installed apps, or ``None`` if it could not be determined

        """
        if installed_apps_response is not None:
            return [
                line.strip().rsplit("package:", 1)[-1] for line in installed_apps_response.splitlines() if line.strip()
            ]

        return None

    @staticmethod
    def _is_volume_muted(stream_music):
        """Determine whether or not the volume is muted from the ``STREAM_MUSIC`` block from ``adb shell dumpsys audio``.

        Parameters
        ----------
        stream_music : str, None
            The ``STREAM_MUSIC`` block from ``adb shell dumpsys audio``

        Returns
        -------
        bool, None
            Whether or not the volume is muted, or ``None`` if it could not be determined

        """
        if not stream_music:
            return None

        matches = re.findall(constants.MUTED_REGEX_PATTERN, stream_music, re.DOTALL | re.MULTILINE)
        if matches:
            return matches[0] == "true"

        return None

    @staticmethod
    def _parse_stream_music(stream_music_raw):
        """Parse the output of the command :py:const:`androidtv.constants.CMD_STREAM_MUSIC`.

        Parameters
        ----------
        stream_music_raw : str, None
            The output of the command :py:const:`androidtv.constants.CMD_STREAM_MUSIC`

        Returns
        -------
        str, None
            The ``STREAM_MUSIC`` block from the output of :py:const:`androidtv.constants.CMD_STREAM_MUSIC`, or ``None`` if it could not be determined

        """
        if not stream_music_raw:
            return None

        matches = re.findall(
            constants.STREAM_MUSIC_REGEX_PATTERN,
            stream_music_raw,
            re.DOTALL | re.MULTILINE,
        )
        if matches:
            return matches[0]

        return None

    @staticmethod
    def _running_apps(running_apps_response):
        """Get the running apps from the output of :py:const:`androidtv.constants.CMD_RUNNING_APPS`.

        Parameters
        ----------
        running_apps_response : str, None
            The output of :py:const:`androidtv.constants.CMD_RUNNING_APPS`

        Returns
        -------
        list, None
            A list of the running apps, or ``None`` if it could not be determined

        """
        if running_apps_response:
            return [line.strip().rsplit(" ", 1)[-1] for line in running_apps_response.splitlines() if line.strip()]

        return None

    @staticmethod
    def _screen_on_awake_wake_lock_size(output):
        """Check if the screen is on and the device is awake, and get the wake lock size.

        Parameters
        ----------
        output : str, None
            The output from :py:const:`androidtv.constants.CMD_SCREEN_ON_AWAKE_WAKE_LOCK_SIZE`

        Returns
        -------
        bool, None
            Whether or not the device is on, or ``None`` if it could not be determined
        bool, None
            Whether or not the device is awake (screensaver is not running), or ``None`` if it could not be determined
        int, None
            The size of the current wake lock, or ``None`` if it could not be determined

        """
        if output is None:
            return None, None, None

        if output == "":
            return False, False, None

        screen_on = output[0] == "1"
        awake = None if len(output) < 2 else output[1] == "1"
        wake_lock_size = None if len(output) < 3 else BaseTV._wake_lock_size(output[2:])

        return screen_on, awake, wake_lock_size

    def _volume(self, stream_music, audio_output_device):
        """Get the absolute volume level from the ``STREAM_MUSIC`` block from ``adb shell dumpsys audio``.

        Parameters
        ----------
        stream_music : str, None
            The ``STREAM_MUSIC`` block from ``adb shell dumpsys audio``
        audio_output_device : str, None
            The current audio playback device

        Returns
        -------
        int, None
            The absolute volume level, or ``None`` if it could not be determined

        """
        if not stream_music:
            return None

        if not self.max_volume:
            max_volume_matches = re.findall(
                constants.MAX_VOLUME_REGEX_PATTERN,
                stream_music,
                re.DOTALL | re.MULTILINE,
            )
            if max_volume_matches:
                self.max_volume = float(max_volume_matches[0])

        if not audio_output_device:
            return None

        volume_matches = re.findall(
            audio_output_device + constants.VOLUME_REGEX_PATTERN,
            stream_music,
            re.DOTALL | re.MULTILINE,
        )
        if volume_matches:
            return int(volume_matches[0])

        return None

    def _volume_level(self, volume):
        """Get the relative volume level from the absolute volume level.

        Parameters
        -------
        volume: int, None
            The absolute volume level

        Returns
        -------
        float, None
            The volume level (between 0 and 1), or ``None`` if it could not be determined

        """
        if volume is not None and self.max_volume:
            return volume / self.max_volume

        return None

    @staticmethod
    def _wake_lock_size(wake_lock_size_response):
        """Get the size of the current wake lock from the output of :py:const:`androidtv.constants.CMD_WAKE_LOCK_SIZE`.

        Parameters
        ----------
        wake_lock_size_response : str, None
            The output of :py:const:`androidtv.constants.CMD_WAKE_LOCK_SIZE`

        Returns
        -------
        int, None
            The size of the current wake lock, or ``None`` if it could not be determined

        """
        if wake_lock_size_response:
            wake_lock_size_matches = constants.REGEX_WAKE_LOCK_SIZE.search(wake_lock_size_response)
            if wake_lock_size_matches:
                return int(wake_lock_size_matches.group("size"))

        return None

    @staticmethod
    def _parse_getevent_line(line):
        """Parse a line of the output received in ``learn_sendevent``.

        Parameters
        ----------
        line : str
            A line of output from ``learn_sendevent``

        Returns
        -------
        str
            The properly formatted ``sendevent`` command

        """
        device_name, event_info = line.split(":", 1)
        integers = [int(x, 16) for x in event_info.strip().split()[:3]]
        return "sendevent {} {} {} {}".format(device_name, *integers)


# ======================================================================= #
#                                                                         #
#                    Validate the state detection rules                   #
#                                                                         #
# ======================================================================= #
def state_detection_rules_validator(rules, exc=KeyError):
    """Validate the rules (i.e., the ``state_detection_rules`` value) for a given app ID (i.e., a key in ``state_detection_rules``).

    For each ``rule`` in ``rules``, this function checks that:

    * ``rule`` is a string or a dictionary
    * If ``rule`` is a string:

        * Check that ``rule`` is in :py:const:`~androidtv.constants.VALID_STATES` or :py:const:`~androidtv.constants.VALID_STATE_PROPERTIES`

    * If ``rule`` is a dictionary:

        * Check that each key is in :py:const:`~androidtv.constants.VALID_STATES`
        * Check that each value is a dictionary

            * Check that each key is in :py:const:`~androidtv.constants.VALID_PROPERTIES`
            * Check that each value is of the right type, according to :py:const:`~androidtv.constants.VALID_PROPERTIES_TYPES`

    See :class:`~androidtv.basetv.basetv.BaseTV` for more info about the ``state_detection_rules`` parameter.

    Parameters
    ----------
    rules : list
        A list of the rules that will be used to determine the state
    exc : Exception
        The exception that will be raised if a rule is invalid

    Returns
    -------
    rules : list
        The provided list of rules

    """
    for rule in rules:
        # A rule must be either a string or a dictionary
        if not isinstance(rule, (str, dict)):
            raise exc("Expected a string or a map, got {}".format(type(rule).__name__))

        # If a rule is a string, check that it is valid
        if isinstance(rule, str):
            if rule not in constants.VALID_STATE_PROPERTIES + constants.VALID_STATES:
                raise exc(
                    "Invalid rule '{0}' is not in {1}".format(
                        rule, constants.VALID_STATE_PROPERTIES + constants.VALID_STATES
                    )
                )

        # If a rule is a dictionary, check that it is valid
        else:
            for state, conditions in rule.items():
                # The keys of the dictionary must be valid states
                if state not in constants.VALID_STATES:
                    raise exc("'{0}' is not a valid state for the 'state_detection_rules' parameter".format(state))

                # The values of the dictionary must be dictionaries
                if not isinstance(conditions, dict):
                    raise exc(
                        "Expected a map for entry '{0}' in 'state_detection_rules', got {1}".format(
                            state, type(conditions).__name__
                        )
                    )

                for prop, value in conditions.items():
                    # The keys of the dictionary must be valid properties that can be checked
                    if prop not in constants.VALID_PROPERTIES:
                        raise exc("Invalid property '{0}' is not in {1}".format(prop, constants.VALID_PROPERTIES))

                    # Make sure the value is of the right type
                    if not isinstance(value, constants.VALID_PROPERTIES_TYPES[prop]):
                        raise exc(
                            "Conditional value for property '{0}' must be of type {1}, not {2}".format(
                                prop,
                                constants.VALID_PROPERTIES_TYPES[prop].__name__,
                                type(value).__name__,
                            )
                        )

    return rules



"""Tests for the ClientAsync and AdbDeviceUsbAsync classes."""

import unittest
from unittest.mock import patch

try:
    from adb_shell.transport.usb_transport import UsbTransport
except (ImportError, OSError):
    UsbTransport = None

from androidtv.adb_manager.adb_manager_async import AdbDeviceUsbAsync, ClientAsync

from .async_wrapper import awaiter
from . import patchers


class TestAsyncClientDevice(unittest.TestCase):
    """Test the ``ClientAsync`` and ``DeviceAsync`` classes defined in ``adb_manager_async.py``.

    These tests can be removed once true async support for using an ADB server is available.

    """

    @awaiter
    async def test_async_client_device(self):
        """Test the ClientAsync class."""
        with patch("androidtv.adb_manager.adb_manager_async.Client", patchers.ClientFakeSuccess):
            client = ClientAsync("host", "port")

            device = await client.device("serial")

            with patch("{}.DeviceFake.shell".format(patchers.__name__)):
                await device.shell("test")

            with patch("{}.DeviceFake.push".format(patchers.__name__)):
                await device.push("local_path", "device_path")

            with patch("{}.DeviceFake.pull".format(patchers.__name__)):
                await device.pull("device_path", "local_path")

            with patch("{}.DeviceFake.screencap".format(patchers.__name__)):
                await device.screencap()

    @awaiter
    async def test_async_client_device_fail(self):
        """Test the ClientAsync class when it fails."""
        with patch("androidtv.adb_manager.adb_manager_async.Client", patchers.ClientFakeFail):
            client = ClientAsync("host", "port")

            device = await client.device("serial")

            self.assertFalse(device)


@unittest.skipIf(UsbTransport is None, "UsbTransport cannot be imported")
class TestAsyncUsb(unittest.TestCase):
    """Test the ``AdbDeviceUsbAsync`` class defined in ``adb_manager_async.py``.

    These tests can be removed once true async support for using a USB connection is available.

    """

    @awaiter
    async def test_async_usb(self):
        """Test the AdbDeviceUsbAsync class."""
        with patch("adb_shell.adb_device.UsbTransport.find_adb", return_value=UsbTransport("device", "setting")):
            device = AdbDeviceUsbAsync()

            self.assertFalse(device.available)

            with patch("androidtv.adb_manager.adb_manager_async.AdbDeviceUsb.connect") as connect:
                await device.connect()
                assert connect.called

            with patch("androidtv.adb_manager.adb_manager_async.AdbDeviceUsb.shell") as shell:
                await device.shell("test")
                assert shell.called

            with patch("androidtv.adb_manager.adb_manager_async.AdbDeviceUsb.push") as push:
                await device.push("local_path", "device_path")
                assert push.called

            with patch("androidtv.adb_manager.adb_manager_async.AdbDeviceUsb.pull") as pull:
                await device.pull("device_path", "local_path")
                assert pull.called

            with patch("androidtv.adb_manager.adb_manager_async.AdbDeviceUsb.close") as close:
                await device.close()
                assert close.called



import sys
import unittest


sys.path.insert(0, "..")

from androidtv import constants
from androidtv.androidtv.base_androidtv import BaseAndroidTV
from androidtv.firetv.base_firetv import BaseFireTV


class TestBaseTV(unittest.TestCase):
    def test_base_android_tv(self):
        """Test that ``BaseAndroidTV.__init__`` runs without error."""
        BaseAndroidTV("host")

    def test_base_fire_tv(self):
        """Test that ``BaseFireTV.__init__`` runs without error."""
        BaseFireTV("host")



import sys
import unittest

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


sys.path.insert(0, "..")

from androidtv import constants, ha_state_detection_rules_validator
from androidtv.androidtv.androidtv_sync import AndroidTVSync
from androidtv.basetv.basetv_sync import BaseTVSync
from . import patchers


DEVICE_PROPERTIES_OUTPUT1 = """Amazon
AFTT
SERIALNO
5.1.1
"""

WIFIMAC_OUTPUT1 = "    link/ether ab:cd:ef:gh:ij:kl brd ff:ff:ff:ff:ff:ff"
ETHMAC_OUTPUT1 = 'Device "eth0" does not exist.'

DEVICE_PROPERTIES_DICT1 = {
    "manufacturer": "Amazon",
    "model": "AFTT",
    "serialno": "SERIALNO",
    "sw_version": "5.1.1",
    "wifimac": "ab:cd:ef:gh:ij:kl",
    "ethmac": None,
}

DEVICE_PROPERTIES_OUTPUT2 = """Amazon
AFTT

5.1.1
"""

DEVICE_PROPERTIES_DICT2 = {
    "manufacturer": "Amazon",
    "model": "AFTT",
    "serialno": None,
    "sw_version": "5.1.1",
    "wifimac": "ab:cd:ef:gh:ij:kl",
    "ethmac": None,
}

DEVICE_PROPERTIES_OUTPUT3 = """Not Amazon
AFTT
SERIALNO
5.1.1
"""

WIFIMAC_OUTPUT3 = 'Device "wlan0" does not exist.'
ETHMAC_OUTPUT3 = "    link/ether ab:cd:ef:gh:ij:kl brd ff:ff:ff:ff:ff:ff"

DEVICE_PROPERTIES_DICT3 = {
    "manufacturer": "Not Amazon",
    "model": "AFTT",
    "serialno": "SERIALNO",
    "sw_version": "5.1.1",
    "wifimac": None,
    "ethmac": "ab:cd:ef:gh:ij:kl",
}


# Source: https://community.home-assistant.io/t/new-chromecast-w-android-tv-integration-only-showing-as-off-or-idle/234424/15
DEVICE_PROPERTIES_GOOGLE_TV = """Google
Chromecast
SERIALNO
10
"""

WIFIMAC_GOOGLE = "    link/ether ab:cd:ef:gh:ij:kl brd ff:ff:ff:ff:ff:ff"
ETHMAC_GOOGLE = 'Device "eth0" does not exist.'

DEVICE_PROPERTIES_OUTPUT_SONY_TV = """Sony
BRAVIA 4K GB
SERIALNO
8.0.0
"""

WIFIMAC_SONY = "    link/ether 11:22:33:44:55:66 brd ff:ff:ff:ff:ff:ff"
ETHMAC_SONY = "    link/ether ab:cd:ef:gh:ij:kl brd ff:ff:ff:ff:ff:ff"

DEVICE_PROPERTIES_DICT_SONY_TV = {
    "manufacturer": "Sony",
    "model": "BRAVIA 4K GB",
    "serialno": "SERIALNO",
    "sw_version": "8.0.0",
    "wifimac": "11:22:33:44:55:66",
    "ethmac": "ab:cd:ef:gh:ij:kl",
}

DEVICE_PROPERTIES_OUTPUT_SHIELD_TV_11 = """NVIDIA
SHIELD Android TV
0123456789012
11
"""

WIFIMAC_SHIELD_TV_11 = "    link/ether 11:22:33:44:55:66 brd ff:ff:ff:ff:ff:ff"
ETHMAC_SHIELD_TV_11 = "    link/ether ab:cd:ef:gh:ij:kl brd ff:ff:ff:ff:ff:ff"

DEVICE_PROPERTIES_OUTPUT_SHIELD_TV_12 = """NVIDIA
SHIELD Android TV
0123456789012
12
"""

WIFIMAC_SHIELD_TV_12 = "    link/ether 11:22:33:44:55:66 brd ff:ff:ff:ff:ff:ff"
ETHMAC_SHIELD_TV_12 = "    link/ether ab:cd:ef:gh:ij:kl brd ff:ff:ff:ff:ff:ff"

DEVICE_PROPERTIES_OUTPUT_SHIELD_TV_13 = """NVIDIA
SHIELD Android TV
0123456789012
13
"""

WIFIMAC_SHIELD_TV_13 = "    link/ether 11:22:33:44:55:66 brd ff:ff:ff:ff:ff:ff"
ETHMAC_SHIELD_TV_13 = "    link/ether ab:cd:ef:gh:ij:kl brd ff:ff:ff:ff:ff:ff"

DEVICE_PROPERTIES_DICT_SHIELD_TV_11 = {
    "manufacturer": "NVIDIA",
    "model": "SHIELD Android TV",
    "serialno": "0123456789012",
    "sw_version": "11",
    "wifimac": "11:22:33:44:55:66",
    "ethmac": "ab:cd:ef:gh:ij:kl",
}

DEVICE_PROPERTIES_DICT_SHIELD_TV_12 = {
    "manufacturer": "NVIDIA",
    "model": "SHIELD Android TV",
    "serialno": "0123456789012",
    "sw_version": "12",
    "wifimac": "11:22:33:44:55:66",
    "ethmac": "ab:cd:ef:gh:ij:kl",
}

DEVICE_PROPERTIES_DICT_SHIELD_TV_13 = {
    "manufacturer": "NVIDIA",
    "model": "SHIELD Android TV",
    "serialno": "0123456789012",
    "sw_version": "13",
    "wifimac": "11:22:33:44:55:66",
    "ethmac": "ab:cd:ef:gh:ij:kl",
}

INSTALLED_APPS_OUTPUT_1 = """package:org.example.app
package:org.example.launcher
"""

INSTALLED_APPS_LIST = ["org.example.app", "org.example.launcher"]

MEDIA_SESSION_STATE_OUTPUT = "com.amazon.tv.launcher\nstate=PlaybackState {state=2, position=0, buffered position=0, speed=0.0, updated=65749, actions=240640, custom actions=[], active item id=-1, error=null}"

STATE_DETECTION_RULES1 = {"com.amazon.tv.launcher": ["off"]}
STATE_DETECTION_RULES2 = {"com.amazon.tv.launcher": ["media_session_state", "off"]}
STATE_DETECTION_RULES3 = {"com.amazon.tv.launcher": [{"standby": {"wake_lock_size": 2}}]}
STATE_DETECTION_RULES4 = {"com.amazon.tv.launcher": [{"standby": {"wake_lock_size": 1}}, "paused"]}
STATE_DETECTION_RULES5 = {"com.amazon.tv.launcher": ["audio_state"]}

STATE_DETECTION_RULES_INVALID1 = {123: ["media_session_state"]}
STATE_DETECTION_RULES_INVALID2 = {"com.amazon.tv.launcher": [123]}
STATE_DETECTION_RULES_INVALID3 = {"com.amazon.tv.launcher": ["INVALID"]}
STATE_DETECTION_RULES_INVALID4 = {"com.amazon.tv.launcher": [{"INVALID": {"wake_lock_size": 2}}]}
STATE_DETECTION_RULES_INVALID5 = {"com.amazon.tv.launcher": [{"standby": "INVALID"}]}
STATE_DETECTION_RULES_INVALID6 = {"com.amazon.tv.launcher": [{"standby": {"INVALID": 2}}]}
STATE_DETECTION_RULES_INVALID7 = {"com.amazon.tv.launcher": [{"standby": {"wake_lock_size": "INVALID"}}]}
STATE_DETECTION_RULES_INVALID8 = {"com.amazon.tv.launcher": [{"standby": {"media_session_state": "INVALID"}}]}
STATE_DETECTION_RULES_INVALID9 = {"com.amazon.tv.launcher": [{"standby": {"audio_state": 123}}]}
STATE_DETECTION_RULES_INVALID10 = {"com.amazon.tv.launcher": [{"standby": {"media_session_state": "INVALID"}}]}
STATE_DETECTION_RULES_INVALID11 = {"com.amazon.tv.launcher": [{"standby": {"audio_state": 123}}]}

PNG_IMAGE = b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\n\x00\x00\x00\n\x08\x06\x00\x00\x00\x8d2\xcf\xbd\x00\x00\x00\x04sBIT\x08\x08\x08\x08|\x08d\x88\x00\x00\x00\tpHYs\x00\x00\x0fa\x00\x00\x0fa\x01\xa8?\xa7i\x00\x00\x00\x0eIDAT\x18\x95c`\x18\x05\x83\x13\x00\x00\x01\x9a\x00\x01\x16\xca\xd3i\x00\x00\x00\x00IEND\xaeB`\x82"


class TestBaseTVSyncPython(unittest.TestCase):
    PATCH_KEY = "python"
    ADB_ATTR = "_adb"

    def setUp(self):
        with patchers.PATCH_ADB_DEVICE_TCP, patchers.patch_connect(True)[self.PATCH_KEY], patchers.patch_shell("")[
            self.PATCH_KEY
        ]:
            self.btv = BaseTVSync("HOST", 5555)
            self.btv.adb_connect()

    def test_available(self):
        """Test that the available property works correctly."""
        self.assertTrue(self.btv.available)

    def test_adb_close(self):
        """Test that the ``adb_close`` method works correctly."""
        self.btv.adb_close()
        if self.PATCH_KEY == "python":
            self.assertFalse(self.btv.available)
        else:
            self.assertTrue(self.btv.available)

    def test_adb_pull(self):
        """Test that the ``adb_pull`` method works correctly."""
        with patchers.PATCH_PULL[self.PATCH_KEY] as patch_pull:
            self.btv.adb_pull("TEST_LOCAL_PATCH", "TEST_DEVICE_PATH")
            self.assertEqual(patch_pull.call_count, 1)

    def test_adb_push(self):
        """Test that the ``adb_push`` method works correctly."""
        with patchers.PATCH_PUSH[self.PATCH_KEY] as patch_push:
            self.btv.adb_push("TEST_LOCAL_PATCH", "TEST_DEVICE_PATH")
            self.assertEqual(patch_push.call_count, 1)

    def test_adb_screencap(self):
        """Test that the ``adb_screencap`` method works correctly."""
        with patch.object(self.btv._adb, "screencap", return_value=PNG_IMAGE):
            self.assertEqual(self.btv.adb_screencap(), PNG_IMAGE)

    def test_keys(self):
        """Test that the key methods send the correct commands."""
        with patchers.patch_connect(True)[self.PATCH_KEY], patchers.patch_shell("")[self.PATCH_KEY]:
            self.btv.adb_shell("TEST")
            self.assertEqual(getattr(self.btv._adb, self.ADB_ATTR).shell_cmd, "TEST")

            self.btv.space()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_SPACE),
            )

            self.btv.key_0()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_0),
            )

            self.btv.key_1()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_1),
            )

            self.btv.key_2()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_2),
            )

            self.btv.key_3()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_3),
            )

            self.btv.key_4()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_4),
            )

            self.btv.key_5()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_5),
            )

            self.btv.key_6()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_6),
            )

            self.btv.key_7()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_7),
            )

            self.btv.key_8()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_8),
            )

            self.btv.key_9()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_9),
            )

            self.btv.key_a()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_A),
            )

            self.btv.key_b()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_B),
            )

            self.btv.key_c()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_C),
            )

            self.btv.key_d()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_D),
            )

            self.btv.key_e()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_E),
            )

            self.btv.key_f()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_F),
            )

            self.btv.key_g()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_G),
            )

            self.btv.key_h()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_H),
            )

            self.btv.key_i()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_I),
            )

            self.btv.key_j()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_J),
            )

            self.btv.key_k()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_K),
            )

            self.btv.key_l()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_L),
            )

            self.btv.key_m()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_M),
            )

            self.btv.key_n()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_N),
            )

            self.btv.key_o()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_O),
            )

            self.btv.key_p()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_P),
            )

            self.btv.key_q()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_Q),
            )

            self.btv.key_r()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_R),
            )

            self.btv.key_s()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_S),
            )

            self.btv.key_t()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_T),
            )

            self.btv.key_u()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_U),
            )

            self.btv.key_v()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_V),
            )

            self.btv.key_w()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_W),
            )

            self.btv.key_x()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_X),
            )

            self.btv.key_y()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_Y),
            )

            self.btv.key_z()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_Z),
            )

            self.btv.power()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_POWER),
            )

            self.btv.sleep()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_SLEEP),
            )

            self.btv.home()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_HOME),
            )

            self.btv.up()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_UP),
            )

            self.btv.down()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_DOWN),
            )

            self.btv.left()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_LEFT),
            )

            self.btv.right()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_RIGHT),
            )

            self.btv.enter()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_ENTER),
            )

            self.btv.back()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_BACK),
            )

            self.btv.menu()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_MENU),
            )

            self.btv.mute_volume()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_MUTE),
            )

            self.btv.media_play()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_PLAY),
            )

            self.btv.media_pause()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_PAUSE),
            )

            self.btv.media_play_pause()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_PLAY_PAUSE),
            )

            self.btv.media_stop()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_STOP),
            )

            self.btv.media_next_track()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_NEXT),
            )

            self.btv.media_previous_track()
            self.assertEqual(
                getattr(self.btv._adb, self.ADB_ATTR).shell_cmd,
                "input keyevent {}".format(constants.KEY_PREVIOUS),
            )

    def test_get_device_properties(self):
        """Check that ``get_device_properties`` works correctly."""
        with patch.object(
            self.btv._adb,
            "shell",
            side_effect=(DEVICE_PROPERTIES_OUTPUT1, ETHMAC_OUTPUT1, WIFIMAC_OUTPUT1),
        ):
            device_properties = self.btv.get_device_properties()
            self.assertDictEqual(DEVICE_PROPERTIES_DICT1, device_properties)

        with patch.object(
            self.btv._adb,
            "shell",
            side_effect=(DEVICE_PROPERTIES_OUTPUT2, ETHMAC_OUTPUT1, WIFIMAC_OUTPUT1),
        ):
            device_properties = self.btv.get_device_properties()
            self.assertDictEqual(DEVICE_PROPERTIES_DICT2, device_properties)

        with patch.object(
            self.btv._adb,
            "shell",
            side_effect=(DEVICE_PROPERTIES_OUTPUT3, ETHMAC_OUTPUT3, WIFIMAC_OUTPUT3),
        ):
            device_properties = self.btv.get_device_properties()
            self.assertDictEqual(DEVICE_PROPERTIES_DICT3, device_properties)

        with patch.object(self.btv._adb, "shell", side_effect=("manufacturer", None, "No match")):
            device_properties = self.btv.get_device_properties()
            self.assertDictEqual({"ethmac": None, "wifimac": None}, device_properties)

        with patch.object(self.btv._adb, "shell", side_effect=("", None, "No match")):
            device_properties = self.btv.get_device_properties()
            self.assertDictEqual({"ethmac": None, "wifimac": None}, device_properties)

        with patch.object(
            self.btv._adb,
            "shell",
            side_effect=(DEVICE_PROPERTIES_GOOGLE_TV, ETHMAC_GOOGLE, WIFIMAC_GOOGLE),
        ):
            self.btv = AndroidTVSync.from_base(self.btv)
            device_properties = self.btv.get_device_properties()
            assert "Chromecast" in self.btv.device_properties.get("model", "")
            assert self.btv.DEVICE_ENUM == AndroidTVSync.DEVICE_ENUM
            self.assertEqual(self.btv.device_properties["manufacturer"], "Google")
            self.assertEqual(self.btv._cmd_current_app(), constants.CMD_CURRENT_APP_GOOGLE_TV)
            self.assertEqual(
                self.btv._cmd_current_app_media_session_state(),
                constants.CMD_CURRENT_APP_MEDIA_SESSION_STATE_GOOGLE_TV,
            )
            self.assertEqual(
                self.btv._cmd_launch_app("TEST"),
                constants.CMD_LAUNCH_APP_GOOGLE_TV.format("TEST"),
            )

        with patch.object(
            self.btv._adb,
            "shell",
            side_effect=(DEVICE_PROPERTIES_OUTPUT_SONY_TV, ETHMAC_SONY, WIFIMAC_SONY),
        ):
            device_properties = self.btv.get_device_properties()
            self.assertDictEqual(DEVICE_PROPERTIES_DICT_SONY_TV, device_properties)

        with patch.object(
            self.btv._adb,
            "shell",
            side_effect=(
                DEVICE_PROPERTIES_OUTPUT_SHIELD_TV_11,
                ETHMAC_SHIELD_TV_11,
                WIFIMAC_SHIELD_TV_11,
            ),
        ):
            self.btv = AndroidTVSync.from_base(self.btv)
            device_properties = self.btv.get_device_properties()
            assert self.btv.device_properties.get("sw_version", "") == "11"
            assert self.btv.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV
            self.assertDictEqual(DEVICE_PROPERTIES_DICT_SHIELD_TV_11, device_properties)
            # _cmd_audio_state
            self.assertEqual(self.btv._cmd_audio_state(), constants.CMD_AUDIO_STATE11)
            # _cmd_volume_set
            self.assertEqual(self.btv._cmd_volume_set(5), constants.CMD_VOLUME_SET_COMMAND11.format(5))
            # _cmd_current_app
            self.assertEqual(self.btv._cmd_current_app(), constants.CMD_CURRENT_APP11)
            # _cmd_current_app_media_session_state
            self.assertEqual(
                self.btv._cmd_current_app_media_session_state(),
                constants.CMD_CURRENT_APP_MEDIA_SESSION_STATE11,
            )
            # _cmd_hdmi_input
            self.assertEqual(self.btv._cmd_hdmi_input(), constants.CMD_HDMI_INPUT11)
            # _cmd_launch_app
            self.assertEqual(
                self.btv._cmd_launch_app("TEST"),
                constants.CMD_LAUNCH_APP11.format("TEST"),
            )

        with patch.object(
            self.btv._adb,
            "shell",
            side_effect=(
                DEVICE_PROPERTIES_OUTPUT_SHIELD_TV_12,
                ETHMAC_SHIELD_TV_12,
                WIFIMAC_SHIELD_TV_12,
            ),
        ):
            self.btv = AndroidTVSync.from_base(self.btv)
            device_properties = self.btv.get_device_properties()
            assert self.btv.device_properties.get("sw_version", "") == "12"
            assert self.btv.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV
            self.assertDictEqual(DEVICE_PROPERTIES_DICT_SHIELD_TV_12, device_properties)
            # _cmd_audio_state
            self.assertEqual(self.btv._cmd_audio_state(), constants.CMD_AUDIO_STATE11)
            # _cmd_volume_set
            self.assertEqual(self.btv._cmd_volume_set(5), constants.CMD_VOLUME_SET_COMMAND11.format(5))
            # _cmd_current_app
            self.assertEqual(self.btv._cmd_current_app(), constants.CMD_CURRENT_APP12)
            # _cmd_current_app_media_session_state
            self.assertEqual(
                self.btv._cmd_current_app_media_session_state(),
                constants.CMD_CURRENT_APP_MEDIA_SESSION_STATE12,
            )
            # _cmd_hdmi_input
            self.assertEqual(self.btv._cmd_hdmi_input(), constants.CMD_HDMI_INPUT11)
            # _cmd_launch_app
            self.assertEqual(
                self.btv._cmd_launch_app("TEST"),
                constants.CMD_LAUNCH_APP12.format("TEST"),
            )

        with patch.object(
            self.btv._adb,
            "shell",
            side_effect=(
                DEVICE_PROPERTIES_OUTPUT_SHIELD_TV_13,
                ETHMAC_SHIELD_TV_13,
                WIFIMAC_SHIELD_TV_13,
            ),
        ):
            self.btv = AndroidTVSync.from_base(self.btv)
            device_properties = self.btv.get_device_properties()
            assert self.btv.device_properties.get("sw_version", "") == "13"
            assert self.btv.DEVICE_ENUM == constants.DeviceEnum.ANDROIDTV
            self.assertDictEqual(DEVICE_PROPERTIES_DICT_SHIELD_TV_13, device_properties)
            # _cmd_audio_state
            self.assertEqual(self.btv._cmd_audio_state(), constants.CMD_AUDIO_STATE11)
            # _cmd_volume_set
            self.assertEqual(self.btv._cmd_volume_set(5), constants.CMD_VOLUME_SET_COMMAND11.format(5))
            # _cmd_current_app
            self.assertEqual(self.btv._cmd_current_app(), constants.CMD_CURRENT_APP13)
            # _cmd_current_app_media_session_state
            self.assertEqual(
                self.btv._cmd_current_app_media_session_state(),
                constants.CMD_CURRENT_APP_MEDIA_SESSION_STATE13,
            )
            # _cmd_hdmi_input
            self.assertEqual(self.btv._cmd_hdmi_input(), constants.CMD_HDMI_INPUT11)
            # _cmd_launch_app
            self.assertEqual(
                self.btv._cmd_launch_app("TEST"),
                constants.CMD_LAUNCH_APP13.format("TEST"),
            )

    def test_get_installed_apps(self):
        """ "Check that `get_installed_apps` works correctly."""
        with patchers.patch_shell(INSTALLED_APPS_OUTPUT_1)[self.PATCH_KEY]:
            installed_apps = self.btv.get_installed_apps()
            self.assertListEqual(INSTALLED_APPS_LIST, installed_apps)

        with patchers.patch_shell(None)[self.PATCH_KEY]:
            installed_apps = self.btv.get_installed_apps()
            self.assertEqual(None, installed_apps)

    def test_awake(self):
        """Check that the ``awake`` property works correctly."""
        with patchers.patch_shell(None)[self.PATCH_KEY]:
            self.assertFalse(self.btv.awake())

        with patchers.patch_shell("0")[self.PATCH_KEY]:
            self.assertFalse(self.btv.awake())

        with patchers.patch_shell("1")[self.PATCH_KEY]:
            self.assertTrue(self.btv.awake())

    def test_audio_state(self):
        """Check that the ``audio_state`` property works correctly."""
        with patchers.patch_shell(None)[self.PATCH_KEY]:
            audio_state = self.btv.audio_state()
            self.assertIsNone(audio_state, None)

        with patchers.patch_shell("0")[self.PATCH_KEY]:
            audio_state = self.btv.audio_state()
            self.assertEqual(audio_state, constants.STATE_IDLE)

        with patchers.patch_shell("1")[self.PATCH_KEY]:
            audio_state = self.btv.audio_state()
            self.assertEqual(audio_state, constants.STATE_PAUSED)

        with patchers.patch_shell("2")[self.PATCH_KEY]:
            audio_state = self.btv.audio_state()
            self.assertEqual(audio_state, constants.STATE_PLAYING)

    def test_current_app(self):
        """Check that the ``current_app`` property works correctly."""
        with patchers.patch_shell(None)[self.PATCH_KEY]:
            current_app = self.btv.current_app()
            self.assertIsNone(current_app, None)

        with patchers.patch_shell("")[self.PATCH_KEY]:
            current_app = self.btv.current_app()
            self.assertIsNone(current_app, None)

        with patchers.patch_shell("com.amazon.tv.launcher")[self.PATCH_KEY]:
            current_app = self.btv.current_app()
            self.assertEqual(current_app, "com.amazon.tv.launcher")

    def test_media_session_state(self):
        """Check that the ``media_session_state`` property works correctly."""
        with patchers.patch_shell(None)[self.PATCH_KEY]:
            media_session_state = self.btv.media_session_state()
            self.assertIsNone(media_session_state, None)

        with patchers.patch_shell("")[self.PATCH_KEY]:
            media_session_state = self.btv.media_session_state()
            self.assertIsNone(media_session_state, None)

        with patchers.patch_shell("unknown.app\n ")[self.PATCH_KEY]:
            media_session_state = self.btv.media_session_state()
            self.assertIsNone(media_session_state, None)

        with patchers.patch_shell("2")[self.PATCH_KEY]:
            media_session_state = self.btv.media_session_state()
            self.assertIsNone(media_session_state, None)

        with patchers.patch_shell(MEDIA_SESSION_STATE_OUTPUT)[self.PATCH_KEY]:
            media_session_state = self.btv.media_session_state()
            self.assertEqual(media_session_state, 2)

    def test_screen_on(self):
        """Check that the ``screen_on`` property works correctly."""
        with patchers.patch_shell(None)[self.PATCH_KEY]:
            self.assertFalse(self.btv.screen_on())

        with patchers.patch_shell("0")[self.PATCH_KEY]:
            self.assertFalse(self.btv.screen_on())

        with patchers.patch_shell("1")[self.PATCH_KEY]:
            self.assertTrue(self.btv.screen_on())

    def test_screen_on_awake_wake_lock_size(self):
        """Check that the ``screen_on_awake_wake_lock_size`` property works correctly."""
        with patchers.patch_shell(None)[self.PATCH_KEY]:
            self.assertTupleEqual(self.btv.screen_on_awake_wake_lock_size(), (None, None, None))

        with patchers.patch_shell("")[self.PATCH_KEY]:
            self.assertTupleEqual(self.btv.screen_on_awake_wake_lock_size(), (False, False, None))

        with patchers.patch_shell("1")[self.PATCH_KEY]:
            self.assertTupleEqual(self.btv.screen_on_awake_wake_lock_size(), (True, None, None))

        with patchers.patch_shell("11")[self.PATCH_KEY]:
            self.assertTupleEqual(self.btv.screen_on_awake_wake_lock_size(), (True, True, None))

        with patchers.patch_shell("11Wake Locks: size=2")[self.PATCH_KEY]:
            self.assertTupleEqual(self.btv.screen_on_awake_wake_lock_size(), (True, True, 2))

        with patchers.patch_shell(
            [
                "Failed to write while dumping serviceWake Locks: size=2",
                "11Wake Locks: size=2",
            ]
        )[self.PATCH_KEY]:
            self.assertTupleEqual(self.btv.screen_on_awake_wake_lock_size(), (True, True, 2))

    def test_state_detection_rules_validator(self):
        """Check that the ``state_detection_rules_validator`` function works correctly."""
        with patchers.patch_connect(True)["python"], patchers.patch_shell("")["python"]:
            # Make sure that no error is raised when the state detection rules are valid
            BaseTVSync("HOST", 5555, state_detection_rules=STATE_DETECTION_RULES1)
            BaseTVSync("HOST", 5555, state_detection_rules=STATE_DETECTION_RULES2)
            BaseTVSync("HOST", 5555, state_detection_rules=STATE_DETECTION_RULES3)
            BaseTVSync("HOST", 5555, state_detection_rules=STATE_DETECTION_RULES4)
            BaseTVSync("HOST", 5555, state_detection_rules=STATE_DETECTION_RULES5)

            # Make sure that an error is raised when the state detection rules are invalid
            self.assertRaises(
                TypeError,
                BaseTVSync,
                "HOST",
                5555,
                "",
                "",
                5037,
                state_detection_rules=STATE_DETECTION_RULES_INVALID1,
            )
            self.assertRaises(
                KeyError,
                BaseTVSync,
                "HOST",
                5555,
                "",
                "",
                5037,
                state_detection_rules=STATE_DETECTION_RULES_INVALID2,
            )
            self.assertRaises(
                KeyError,
                BaseTVSync,
                "HOST",
                5555,
                "",
                "",
                5037,
                state_detection_rules=STATE_DETECTION_RULES_INVALID3,
            )
            self.assertRaises(
                KeyError,
                BaseTVSync,
                "HOST",
                5555,
                "",
                "",
                5037,
                state_detection_rules=STATE_DETECTION_RULES_INVALID4,
            )
            self.assertRaises(
                KeyError,
                BaseTVSync,
                "HOST",
                5555,
                "",
                "",
                5037,
                state_detection_rules=STATE_DETECTION_RULES_INVALID5,
            )
            self.assertRaises(
                KeyError,
                BaseTVSync,
                "HOST",
                5555,
                "",
                "",
                5037,
                state_detection_rules=STATE_DETECTION_RULES_INVALID6,
            )
            self.assertRaises(
                KeyError,
                BaseTVSync,
                "HOST",
                5555,
                "",
                "",
                5037,
                state_detection_rules=STATE_DETECTION_RULES_INVALID7,
            )
            self.assertRaises(
                KeyError,
                BaseTVSync,
                "HOST",
                5555,
                "",
                "",
                5037,
                state_detection_rules=STATE_DETECTION_RULES_INVALID8,
            )
            self.assertRaises(
                KeyError,
                BaseTVSync,
                "HOST",
                5555,
                "",
                "",
                5037,
                state_detection_rules=STATE_DETECTION_RULES_INVALID9,
            )
            self.assertRaises(
                KeyError,
                BaseTVSync,
                "HOST",
                5555,
                "",
                "",
                5037,
                state_detection_rules=STATE_DETECTION_RULES_INVALID10,
            )
            self.assertRaises(
                KeyError,
                BaseTVSync,
                "HOST",
                5555,
                "",
                "",
                5037,
                state_detection_rules=STATE_DETECTION_RULES_INVALID11,
            )

    def test_wake_lock_size(self):
        """Check that the ``wake_lock_size`` property works correctly."""
        with patchers.patch_shell(None)[self.PATCH_KEY]:
            self.assertIsNone(self.btv.wake_lock_size())

        with patchers.patch_shell("")[self.PATCH_KEY]:
            self.assertIsNone(self.btv.wake_lock_size())

        with patchers.patch_shell("Wake Locks: size=2")[self.PATCH_KEY]:
            self.assertEqual(self.btv.wake_lock_size(), 2)

        with patchers.patch_shell("INVALID")[self.PATCH_KEY]:
            self.assertIsNone(self.btv.wake_lock_size())

    def test_get_hdmi_input(self):
        """Check that the ``get_hdmi_input`` function works correctly."""
        with patchers.patch_shell("HW2")[self.PATCH_KEY]:
            self.assertEqual(self.btv.get_hdmi_input(), "HW2")

        with patchers.patch_shell("HW2\n")[self.PATCH_KEY]:
            self.assertEqual(self.btv.get_hdmi_input(), "HW2")

        with patchers.patch_shell("HW2\r\n")[self.PATCH_KEY]:
            self.assertEqual(self.btv.get_hdmi_input(), "HW2")

        with patchers.patch_shell("")[self.PATCH_KEY]:
            self.assertIsNone(self.btv.get_hdmi_input())

        with patchers.patch_shell("\r\n")[self.PATCH_KEY]:
            self.assertIsNone(self.btv.get_hdmi_input())

        with patchers.patch_shell(None)[self.PATCH_KEY]:
            self.assertIsNone(self.btv.get_hdmi_input())

    def test_learn_sendevent(self):
        """Check that the ``learn_sendevent`` method works correctly."""
        with patchers.patch_shell(
            'add device 1: /dev/input/event4\r\n  name:     "Amazon Fire TV Remote"\r\nadd device 2: /dev/input/event3\r\n  name:     "kcmouse"\r\ncould not get driver version for /dev/input/mouse0, Not a typewriter\r\nadd device 3: /dev/input/event2\r\n  name:     "amazon_touch"\r\nadd device 4: /dev/input/event1\r\n  name:     "hdmipower"\r\nadd device 5: /dev/input/event0\r\n  name:     "mtk-kpd"\r\ncould not get driver version for /dev/input/mice, Not a typewriter\r\n/dev/input/event4: 0004 0004 00070051\r\n/dev/input/event4: 0001 006c 00000001\r\n/dev/input/event4: 0000 0000 00000000\r\n/dev/input/event4: 0004 0004 00070051\r\n/dev/input/event4: 0001 006c 00000000\r\n/dev/input/event4: 0000 0000 00000000\r\nyour command was interrupted'
        )[self.PATCH_KEY]:
            self.assertEqual(
                self.btv.learn_sendevent(),
                "sendevent /dev/input/event4 4 4 458833 && sendevent /dev/input/event4 1 108 1 && sendevent /dev/input/event4 0 0 0 && sendevent /dev/input/event4 4 4 458833 && sendevent /dev/input/event4 1 108 0 && sendevent /dev/input/event4 0 0 0",
            )

        with patchers.patch_shell("This is not a valid response")[self.PATCH_KEY]:
            self.assertEqual(self.btv.learn_sendevent(), "")


class TestHAStateDetectionRulesValidator(unittest.TestCase):
    def test_ha_state_detection_rules_validator(self):
        """Check that ``ha_state_detection_rules_validator()`` works correctly."""
        with self.assertRaises(AssertionError):
            for app_id, rules in STATE_DETECTION_RULES_INVALID2.items():
                ha_state_detection_rules_validator(AssertionError)(rules)


if __name__ == "__main__":
    unittest.main()



"""Communicate with an Amazon Fire TV device via ADB over a network.

ADB Debugging must be enabled.
"""


import logging

from ..basetv.basetv import BaseTV
from .. import constants

_LOGGER = logging.getLogger(__name__)


class BaseFireTV(BaseTV):  # pylint: disable=too-few-public-methods
    """Representation of an Amazon Fire TV device.

    Parameters
    ----------
    host : str
        The address of the device; may be an IP address or a host name
    port : int
        The device port to which we are connecting (default is 5555)
    adbkey : str
        The path to the ``adbkey`` file for ADB authentication
    adb_server_ip : str
        The IP address of the ADB server
    adb_server_port : int
        The port for the ADB server
    state_detection_rules : dict, None
        A dictionary of rules for determining the state (see :class:`~androidtv.basetv.basetv.BaseTV`)

    """

    DEVICE_CLASS = "firetv"
    DEVICE_ENUM = constants.DeviceEnum.FIRETV

    def __init__(self, host, port=5555, adbkey="", adb_server_ip="", adb_server_port=5037, state_detection_rules=None):
        BaseTV.__init__(self, None, host, port, adbkey, adb_server_ip, adb_server_port, state_detection_rules)

    # ======================================================================= #
    #                                                                         #
    #                          Home Assistant Update                          #
    #                                                                         #
    # ======================================================================= #
    def _update(self, screen_on, awake, wake_lock_size, current_app, media_session_state, running_apps, hdmi_input):
        """Get the info needed for a Home Assistant update.

        Parameters
        ----------
        screen_on : bool, None
            Whether or not the device is on, or ``None`` if it was not determined
        awake : bool, None
            Whether or not the device is awake (screensaver is not running), or ``None`` if it was not determined
        wake_lock_size : int, None
            The size of the current wake lock, or ``None`` if it was not determined
        current_app : str, None
            The current app property, or ``None`` if it was not determined
        media_session_state : int, None
            The state from the output of ``dumpsys media_session``, or ``None`` if it was not determined
        running_apps : list, None
            A list of the running apps, or ``None`` if it was not determined
        hdmi_input : str, None
            The HDMI input, or ``None`` if it could not be determined

        Returns
        -------
        state : str
            The state of the device
        current_app : str
            The current running app
        running_apps : list
            A list of the running apps if ``get_running_apps`` is True, otherwise the list ``[current_app]``
        hdmi_input : str, None
            The HDMI input, or ``None`` if it could not be determined

        """
        # Check if device is unavailable
        if screen_on is None:
            state = None
            current_app = None
            running_apps = None

        # Check if device is off
        elif not screen_on:
            state = constants.STATE_OFF
            current_app = None
            running_apps = None

        # Check if screen saver is on
        elif not awake:
            state = constants.STATE_STANDBY
            current_app = None
            running_apps = None

        else:
            # Get the running apps
            if not running_apps and current_app:
                running_apps = [current_app]

            # Determine the state using custom rules
            state = self._custom_state_detection(
                current_app=current_app, media_session_state=media_session_state, wake_lock_size=wake_lock_size
            )
            if state:
                return state, current_app, running_apps, hdmi_input

            # Determine the state based on the `current_app`
            if current_app in [constants.APP_FIRETV_PACKAGE_LAUNCHER, constants.APP_FIRETV_PACKAGE_SETTINGS, None]:
                state = constants.STATE_IDLE

            # Amazon Video
            elif current_app == constants.APP_AMAZON_VIDEO:
                if media_session_state == 2:
                    state = constants.STATE_PAUSED
                elif media_session_state == 3:
                    state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_IDLE

            # Firefox
            elif current_app == constants.APP_FIREFOX:
                if wake_lock_size == 3:
                    state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_IDLE

            # Hulu
            elif current_app == constants.APP_HULU:
                if wake_lock_size == 4:
                    state = constants.STATE_PLAYING
                elif wake_lock_size == 2:
                    state = constants.STATE_PAUSED
                else:
                    state = constants.STATE_IDLE

            # Jellyfin
            elif current_app == constants.APP_JELLYFIN_TV:
                if wake_lock_size == 2:
                    state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_PAUSED

            # Netflix
            elif current_app == constants.APP_NETFLIX:
                if media_session_state == 2:
                    state = constants.STATE_PAUSED
                elif media_session_state == 3:
                    state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_IDLE

            # Plex
            elif current_app == constants.APP_PLEX:
                if media_session_state == 3:
                    if wake_lock_size == 2:
                        state = constants.STATE_PAUSED
                    else:
                        state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_IDLE

            # Sport 1
            elif current_app == constants.APP_SPORT1:
                if wake_lock_size == 2:
                    state = constants.STATE_PAUSED
                elif wake_lock_size == 3:
                    state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_IDLE

            # Spotify
            elif current_app == constants.APP_SPOTIFY:
                if media_session_state == 2:
                    state = constants.STATE_PAUSED
                elif media_session_state == 3:
                    state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_IDLE

            # RTL Plus (Germany)
            elif current_app == constants.APP_TVNOW:
                if wake_lock_size == 3:
                    state = constants.STATE_PAUSED
                elif wake_lock_size == 4:
                    state = constants.STATE_PLAYING
                elif wake_lock_size == 5:
                    state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_IDLE

            # Twitch
            elif current_app == constants.APP_TWITCH_FIRETV:
                if wake_lock_size == 2:
                    state = constants.STATE_PAUSED
                elif media_session_state == 3:
                    state = constants.STATE_PLAYING
                elif media_session_state == 4:
                    state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_IDLE

            # Waipu TV
            elif current_app == constants.APP_WAIPU_TV:
                if wake_lock_size == 2:
                    state = constants.STATE_PAUSED
                elif wake_lock_size == 3:
                    state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_IDLE

            # Get the state from `media_session_state`
            elif media_session_state:
                if media_session_state == 2:
                    state = constants.STATE_PAUSED
                elif media_session_state == 3:
                    state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_IDLE

            # Get the state from `wake_lock_size`
            else:
                if wake_lock_size == 1:
                    state = constants.STATE_PLAYING
                else:
                    state = constants.STATE_PAUSED

        return state, current_app, running_apps, hdmi_input



import sys
import unittest
from unittest.mock import patch


sys.path.insert(0, "..")

from androidtv.setup_async import setup
from androidtv.androidtv.androidtv_async import AndroidTVAsync
from androidtv.firetv.firetv_async import FireTVAsync

from . import async_patchers
from .async_wrapper import awaiter


DEVICE_PROPERTIES_OUTPUT1 = "Amazon\n\n\n123"

DEVICE_PROPERTIES_DICT1 = {
    "manufacturer": "Amazon",
    "model": "",
    "serialno": None,
    "sw_version": "123",
    "wifimac": None,
    "ethmac": None,
}

DEVICE_PROPERTIES_OUTPUT2 = "Not Amazon\n\n\n456"

DEVICE_PROPERTIES_DICT2 = {
    "manufacturer": "Not Amazon",
    "model": "",
    "serialno": None,
    "sw_version": "456",
    "wifimac": None,
    "ethmac": None,
}


class TestSetup(unittest.TestCase):
    PATCH_KEY = "python"

    @awaiter
    async def test_setup(self):
        """Test that the ``setup`` function works correctly."""
        with self.assertRaises(ValueError):
            await setup("HOST", 5555, device_class="INVALID")

        with async_patchers.PATCH_ADB_DEVICE_TCP, async_patchers.patch_connect(True)[
            self.PATCH_KEY
        ], async_patchers.patch_shell(DEVICE_PROPERTIES_OUTPUT1)[self.PATCH_KEY]:
            ftv = await setup("HOST", 5555)
            self.assertIsInstance(ftv, FireTVAsync)
            self.assertDictEqual(ftv.device_properties, DEVICE_PROPERTIES_DICT1)

        with async_patchers.PATCH_ADB_DEVICE_TCP, async_patchers.patch_connect(True)[
            self.PATCH_KEY
        ], async_patchers.patch_shell(DEVICE_PROPERTIES_OUTPUT2)[self.PATCH_KEY]:
            atv = await setup("HOST", 5555)
            self.assertIsInstance(atv, AndroidTVAsync)
            self.assertDictEqual(atv.device_properties, DEVICE_PROPERTIES_DICT2)

        with async_patchers.PATCH_ADB_DEVICE_TCP, async_patchers.patch_connect(True)[
            self.PATCH_KEY
        ], async_patchers.patch_shell(DEVICE_PROPERTIES_OUTPUT1)[self.PATCH_KEY]:
            ftv = await setup("HOST", 5555, device_class="androidtv")
            self.assertIsInstance(ftv, AndroidTVAsync)
            self.assertDictEqual(ftv.device_properties, DEVICE_PROPERTIES_DICT1)

        with async_patchers.PATCH_ADB_DEVICE_TCP, async_patchers.patch_connect(True)[
            self.PATCH_KEY
        ], async_patchers.patch_shell(DEVICE_PROPERTIES_OUTPUT2)[self.PATCH_KEY]:
            atv = await setup("HOST", 5555, device_class="firetv")
            self.assertIsInstance(atv, FireTVAsync)
            self.assertDictEqual(atv.device_properties, DEVICE_PROPERTIES_DICT2)


if __name__ == "__main__":
    unittest.main()



"""Constants used throughout the code.

**Links**

* `ADB key event codes <https://developer.android.com/reference/android/view/KeyEvent>`_
* `MediaSession PlaybackState property <https://developer.android.com/reference/android/media/session/PlaybackState.html>`_

"""


import re
import sys

if sys.version_info[0] == 3 and sys.version_info[1] >= 5:
    from enum import IntEnum, unique
else:  # pragma: no cover
    IntEnum = object

    def unique(cls):
        """A class decorator that does nothing."""
        return cls


@unique
class DeviceEnum(IntEnum):
    """An enum for the various device types."""

    BASETV = 0
    ANDROIDTV = 1
    FIRETV = 2


# Intents
INTENT_LAUNCH = "android.intent.category.LEANBACK_LAUNCHER"
INTENT_LAUNCH_FIRETV = "android.intent.category.LAUNCHER"
INTENT_HOME = "android.intent.category.HOME"

# Customizable commands
CUSTOM_AUDIO_STATE = "audio_state"
CUSTOM_CURRENT_APP = "current_app"
CUSTOM_CURRENT_APP_MEDIA_SESSION_STATE = "current_app_media_session_state"
CUSTOM_HDMI_INPUT = "hdmi_input"
CUSTOM_LAUNCH_APP = "launch_app"
CUSTOM_RUNNING_APPS = "running_apps"
CUSTOM_TURN_OFF = "turn_off"
CUSTOM_TURN_ON = "turn_on"
CUSTOMIZABLE_COMMANDS = {
    CUSTOM_AUDIO_STATE,
    CUSTOM_CURRENT_APP,
    CUSTOM_CURRENT_APP_MEDIA_SESSION_STATE,
    CUSTOM_HDMI_INPUT,
    CUSTOM_LAUNCH_APP,
    CUSTOM_RUNNING_APPS,
    CUSTOM_TURN_OFF,
    CUSTOM_TURN_ON,
}

#: The subset of `CUSTOMIZABLE_COMMANDS` that is potentially used in the ``update()`` method
HA_CUSTOMIZABLE_COMMANDS = (
    CUSTOM_AUDIO_STATE,
    CUSTOM_CURRENT_APP_MEDIA_SESSION_STATE,
    CUSTOM_HDMI_INPUT,
    CUSTOM_LAUNCH_APP,
    CUSTOM_RUNNING_APPS,
    CUSTOM_TURN_OFF,
    CUSTOM_TURN_ON,
)

# echo '1' if the previous shell command was successful
CMD_SUCCESS1 = r" && echo -e '1\c'"

# echo '1' if the previous shell command was successful, echo '0' if it was not
CMD_SUCCESS1_FAILURE0 = r" && echo -e '1\c' || echo -e '0\c'"

#: Get the audio state
CMD_AUDIO_STATE = r"dumpsys audio | grep paused | grep -qv 'Buffer Queue' && echo -e '1\c' || (dumpsys audio | grep started | grep -qv 'Buffer Queue' && echo '2\c' || echo '0\c')"

#: Get the audio state for an Android 11 device
CMD_AUDIO_STATE11 = (
    "CURRENT_AUDIO_STATE=$(dumpsys audio | sed -r -n '/[0-9]{2}-[0-9]{2}.*player piid:.*state:.*$/h; ${x;p;}') && "
    + r"echo $CURRENT_AUDIO_STATE | grep -q paused && echo -e '1\c' || { echo $CURRENT_AUDIO_STATE | grep -q started && echo '2\c' || echo '0\c' ; }"
)

#: Determine whether the device is awake
CMD_AWAKE = "dumpsys power | grep mWakefulness | grep -q Awake"

#: Parse current application identifier from dumpsys output and assign it to ``CURRENT_APP`` variable (assumes dumpsys output is momentarily set to ``CURRENT_APP`` variable)
CMD_PARSE_CURRENT_APP = "CURRENT_APP=${CURRENT_APP#*ActivityRecord{* * } && CURRENT_APP=${CURRENT_APP#*{* * } && CURRENT_APP=${CURRENT_APP%%/*} && CURRENT_APP=${CURRENT_APP%\\}*}"

#: Parse current application for an Android 11 device
CMD_PARSE_CURRENT_APP11 = "CURRENT_APP=${CURRENT_APP%%/*} && CURRENT_APP=${CURRENT_APP##* }"
#: Assign focused application identifier to ``CURRENT_APP`` variable
CMD_DEFINE_CURRENT_APP_VARIABLE = (
    "CURRENT_APP=$(dumpsys window windows | grep -E 'mCurrentFocus|mFocusedApp') && " + CMD_PARSE_CURRENT_APP
)
#: Assign focused application identifier to ``CURRENT_APP`` variable for an Android 11 device
CMD_DEFINE_CURRENT_APP_VARIABLE11 = (
    "CURRENT_APP=$(dumpsys window windows | grep -E -m 1 'mInputMethod(Input)?Target') && " + CMD_PARSE_CURRENT_APP11
)

#: Assign focused application identifier to ``CURRENT_APP`` variable for an Android 12 device
CMD_DEFINE_CURRENT_APP_VARIABLE12 = (
    "CURRENT_APP=$(dumpsys window windows | grep -E 'mCurrentFocus|mFocusedApp|mObscuringWindow') && "
    + CMD_PARSE_CURRENT_APP11
)

#: Assign focused application identifier to ``CURRENT_APP`` variable for an Android 13 device
CMD_DEFINE_CURRENT_APP_VARIABLE13 = (
    "CURRENT_APP=$(dumpsys window windows | grep -E -m 1 'imeLayeringTarget|imeInputTarget|imeControlTarget') && "
    + CMD_PARSE_CURRENT_APP11
)

#: Output identifier for current/focused application
CMD_CURRENT_APP = CMD_DEFINE_CURRENT_APP_VARIABLE + " && echo $CURRENT_APP"

#: Output identifier for current/focused application for an Android 11 device
CMD_CURRENT_APP11 = CMD_DEFINE_CURRENT_APP_VARIABLE11 + " && echo $CURRENT_APP"

#: Output identifier for current/focused application for an Android 12 device
CMD_CURRENT_APP12 = CMD_DEFINE_CURRENT_APP_VARIABLE12 + " && echo $CURRENT_APP"

#: Output identifier for current/focused application for an Android 13 device
CMD_CURRENT_APP13 = CMD_DEFINE_CURRENT_APP_VARIABLE13 + " && echo $CURRENT_APP"

#: Assign focused application identifier to ``CURRENT_APP`` variable (for a Google TV device)
CMD_DEFINE_CURRENT_APP_VARIABLE_GOOGLE_TV = (
    "CURRENT_APP=$(dumpsys activity a . | grep mResumedActivity) && " + CMD_PARSE_CURRENT_APP
)

#: Output identifier for current/focused application (for a Google TV device)
CMD_CURRENT_APP_GOOGLE_TV = CMD_DEFINE_CURRENT_APP_VARIABLE_GOOGLE_TV + " && echo $CURRENT_APP"

#: set volume
CMD_VOLUME_SET_COMMAND = "media volume --show --stream 3 --set {}"

#: set volume for an Android 11 & 12 & 13 device
CMD_VOLUME_SET_COMMAND11 = "cmd media_session volume --show --stream 3 --set {}"

#: Get the HDMI input
CMD_HDMI_INPUT = (
    "dumpsys activity starter | grep -E -o '(ExternalTv|HDMI)InputService/HW[0-9]' -m 1 | grep -o 'HW[0-9]'"
)

#: Get the HDMI input for an Android 11 device
CMD_HDMI_INPUT11 = (
    "(HDMI=$(dumpsys tv_input | grep 'ResourceClientProfile {.*}' | grep -o -E '(hdmi_port=[0-9]|TV)') && { echo ${HDMI/hdmi_port=/HW} | cut -d' ' -f1 ; }) || "
    + CMD_HDMI_INPUT
)

#: Launch an app if it is not already the current app (assumes the variable ``CURRENT_APP`` has already been set)
CMD_LAUNCH_APP_CONDITION = (
    "if [ $CURRENT_APP != '{0}' ]; then monkey -p {0} -c " + INTENT_LAUNCH + " --pct-syskeys 0 1; fi"
)

#: Launch an app if it is not already the current app (assumes the variable ``CURRENT_APP`` has already been set) on a Fire TV
CMD_LAUNCH_APP_CONDITION_FIRETV = (
    "if [ $CURRENT_APP != '{0}' ]; then monkey -p {0} -c " + INTENT_LAUNCH_FIRETV + " --pct-syskeys 0 1; fi"
)

#: Launch an app if it is not already the current app
CMD_LAUNCH_APP = (
    CMD_DEFINE_CURRENT_APP_VARIABLE.replace("{", "{{").replace("}", "}}") + " && " + CMD_LAUNCH_APP_CONDITION
)

#: Launch an app if it is not already the current app on an Android 11 device
CMD_LAUNCH_APP11 = (
    CMD_DEFINE_CURRENT_APP_VARIABLE11.replace("{", "{{").replace("}", "}}") + " && " + CMD_LAUNCH_APP_CONDITION
)

#: Launch an app if it is not already the current app on an Android 12 device
CMD_LAUNCH_APP12 = (
    CMD_DEFINE_CURRENT_APP_VARIABLE12.replace("{", "{{").replace("}", "}}") + " && " + CMD_LAUNCH_APP_CONDITION
)

#: Launch an app if it is not already the current app on an Android 11 device
CMD_LAUNCH_APP13 = (
    CMD_DEFINE_CURRENT_APP_VARIABLE13.replace("{", "{{").replace("}", "}}") + " && " + CMD_LAUNCH_APP_CONDITION
)

#: Launch an app on a Fire TV device
CMD_LAUNCH_APP_FIRETV = (
    CMD_DEFINE_CURRENT_APP_VARIABLE.replace("{", "{{").replace("}", "}}") + " && " + CMD_LAUNCH_APP_CONDITION_FIRETV
)

#: Launch an app on a Google TV device
CMD_LAUNCH_APP_GOOGLE_TV = (
    CMD_DEFINE_CURRENT_APP_VARIABLE_GOOGLE_TV.replace("{", "{{").replace("}", "}}") + " && " + CMD_LAUNCH_APP_CONDITION
)

#: Get the state from ``dumpsys media_session``; this assumes that the variable ``CURRENT_APP`` has been defined
CMD_MEDIA_SESSION_STATE = "dumpsys media_session | grep -A 100 'Sessions Stack' | grep -A 100 $CURRENT_APP | grep -m 1 'state=PlaybackState {'"

#: Determine the current app and get the state from ``dumpsys media_session``
CMD_CURRENT_APP_MEDIA_SESSION_STATE = CMD_CURRENT_APP + " && " + CMD_MEDIA_SESSION_STATE

#: Determine the current app and get the state from ``dumpsys media_session`` for an Android 11 device
CMD_CURRENT_APP_MEDIA_SESSION_STATE11 = CMD_CURRENT_APP11 + " && " + CMD_MEDIA_SESSION_STATE

#: Determine the current app and get the state from ``dumpsys media_session`` for an Android 12 device
CMD_CURRENT_APP_MEDIA_SESSION_STATE12 = CMD_CURRENT_APP12 + " && " + CMD_MEDIA_SESSION_STATE

#: Determine the current app and get the state from ``dumpsys media_session`` for an Android 13 device
CMD_CURRENT_APP_MEDIA_SESSION_STATE13 = CMD_CURRENT_APP13 + " && " + CMD_MEDIA_SESSION_STATE

#: Determine the current app and get the state from ``dumpsys media_session`` for a Google TV device
CMD_CURRENT_APP_MEDIA_SESSION_STATE_GOOGLE_TV = CMD_CURRENT_APP_GOOGLE_TV + " && " + CMD_MEDIA_SESSION_STATE

#: Get the running apps for an Android TV device
CMD_RUNNING_APPS_ANDROIDTV = "ps -A | grep u0_a"

#: Get the running apps for a Fire TV device
CMD_RUNNING_APPS_FIRETV = "ps | grep u0_a"

#: Get installed apps
CMD_INSTALLED_APPS = "pm list packages"

#: Determine if the device is on
CMD_SCREEN_ON = "(dumpsys power | grep 'Display Power' | grep -q 'state=ON' || dumpsys power | grep -q 'mScreenOn=true' || dumpsys display | grep -q 'mScreenState=ON')"

#: Get the "STREAM_MUSIC" block from ``dumpsys audio``
CMD_STREAM_MUSIC = r"dumpsys audio | grep '\- STREAM_MUSIC:' -A 11"

#: Turn off an Android TV device (note: `KEY_POWER = 26` is defined below)
CMD_TURN_OFF_ANDROIDTV = CMD_SCREEN_ON + " && input keyevent 26"

#: Turn off a Fire TV device (note: `KEY_SLEEP = 223` is defined below)
CMD_TURN_OFF_FIRETV = CMD_SCREEN_ON + " && input keyevent 223"

#: Turn on an Android TV device (note: `KEY_POWER = 26` is defined below)
CMD_TURN_ON_ANDROIDTV = CMD_SCREEN_ON + " || input keyevent 26"

#: Turn on a Fire TV device (note: `KEY_POWER = 26` and `KEY_HOME = 3` are defined below)
CMD_TURN_ON_FIRETV = CMD_SCREEN_ON + " || (input keyevent 26 && input keyevent 3)"

#: Get the wake lock size
CMD_WAKE_LOCK_SIZE = "dumpsys power | grep Locks | grep 'size='"

#: Determine if the device is on, the screen is on, and get the wake lock size
CMD_SCREEN_ON_AWAKE_WAKE_LOCK_SIZE = (
    CMD_SCREEN_ON + CMD_SUCCESS1_FAILURE0 + " && " + CMD_AWAKE + CMD_SUCCESS1_FAILURE0 + " && " + CMD_WAKE_LOCK_SIZE
)

# `getprop` commands
CMD_MANUFACTURER = "getprop ro.product.manufacturer"
CMD_MODEL = "getprop ro.product.model"
CMD_SERIALNO = "getprop ro.serialno"
CMD_VERSION = "getprop ro.build.version.release"

# Commands for getting the MAC address
CMD_MAC_WLAN0 = "ip addr show wlan0 | grep -m 1 ether"
CMD_MAC_ETH0 = "ip addr show eth0 | grep -m 1 ether"

#: The command used for getting the device properties
CMD_DEVICE_PROPERTIES = CMD_MANUFACTURER + " && " + CMD_MODEL + " && " + CMD_SERIALNO + " && " + CMD_VERSION


# ADB key event codes
# https://developer.android.com/reference/android/view/KeyEvent
KEY_BACK = 4
KEY_BLUE = 186
KEY_CENTER = 23
KEY_COMPONENT1 = 249
KEY_COMPONENT2 = 250
KEY_COMPOSITE1 = 247
KEY_COMPOSITE2 = 248
KEY_DOWN = 20
KEY_END = 123
KEY_ENTER = 66
KEY_ESCAPE = 111
KEY_FAST_FORWARD = 90
KEY_GREEN = 184
KEY_HDMI1 = 243
KEY_HDMI2 = 244
KEY_HDMI3 = 245
KEY_HDMI4 = 246
KEY_HOME = 3
KEY_INPUT = 178
KEY_LEFT = 21
KEY_MENU = 82
KEY_MOVE_HOME = 122
KEY_MUTE = 164
KEY_NEXT = 87
KEY_PAIRING = 225
KEY_PAUSE = 127
KEY_PLAY = 126
KEY_PLAY_PAUSE = 85
KEY_POWER = 26
KEY_PREVIOUS = 88
KEY_RED = 183
KEY_RESUME = 224
KEY_REWIND = 89
KEY_RIGHT = 22
KEY_SAT = 237
KEY_SEARCH = 84
KEY_SETTINGS = 176
KEY_SLEEP = 223
KEY_SPACE = 62
KEY_STOP = 86
KEY_SUSPEND = 276
KEY_SYSDOWN = 281
KEY_SYSLEFT = 282
KEY_SYSRIGHT = 283
KEY_SYSUP = 280
KEY_TEXT = 233
KEY_TOP = 122
KEY_UP = 19
KEY_VGA = 251
KEY_VOLUME_DOWN = 25
KEY_VOLUME_UP = 24
KEY_WAKEUP = 224
KEY_YELLOW = 185


# Alphanumeric key event codes
KEY_0 = 7
KEY_1 = 8
KEY_2 = 9
KEY_3 = 10
KEY_4 = 11
KEY_5 = 12
KEY_6 = 13
KEY_7 = 14
KEY_8 = 15
KEY_9 = 16
KEY_A = 29
KEY_B = 30
KEY_C = 31
KEY_D = 32
KEY_E = 33
KEY_F = 34
KEY_G = 35
KEY_H = 36
KEY_I = 37
KEY_J = 38
KEY_K = 39
KEY_L = 40
KEY_M = 41
KEY_N = 42
KEY_O = 43
KEY_P = 44
KEY_Q = 45
KEY_R = 46
KEY_S = 47
KEY_T = 48
KEY_U = 49
KEY_V = 50
KEY_W = 51
KEY_X = 52
KEY_Y = 53
KEY_Z = 54


# Android TV keys
KEYS = {
    "BACK": KEY_BACK,
    "BLUE": KEY_BLUE,
    "CENTER": KEY_CENTER,
    "COMPONENT1": KEY_COMPONENT1,
    "COMPONENT2": KEY_COMPONENT2,
    "COMPOSITE1": KEY_COMPOSITE1,
    "COMPOSITE2": KEY_COMPOSITE2,
    "DOWN": KEY_DOWN,
    "END": KEY_END,
    "ENTER": KEY_ENTER,
    "ESCAPE": KEY_ESCAPE,
    "FAST_FORWARD": KEY_FAST_FORWARD,
    "GREEN": KEY_GREEN,
    "HDMI1": KEY_HDMI1,
    "HDMI2": KEY_HDMI2,
    "HDMI3": KEY_HDMI3,
    "HDMI4": KEY_HDMI4,
    "HOME": KEY_HOME,
    "INPUT": KEY_INPUT,
    "LEFT": KEY_LEFT,
    "MENU": KEY_MENU,
    "MOVE_HOME": KEY_MOVE_HOME,
    "MUTE": KEY_MUTE,
    "PAIRING": KEY_PAIRING,
    "POWER": KEY_POWER,
    "RED": KEY_RED,
    "RESUME": KEY_RESUME,
    "REWIND": KEY_REWIND,
    "RIGHT": KEY_RIGHT,
    "SAT": KEY_SAT,
    "SEARCH": KEY_SEARCH,
    "SETTINGS": KEY_SETTINGS,
    "SLEEP": KEY_SLEEP,
    "SUSPEND": KEY_SUSPEND,
    "SYSDOWN": KEY_SYSDOWN,
    "SYSLEFT": KEY_SYSLEFT,
    "SYSRIGHT": KEY_SYSRIGHT,
    "SYSUP": KEY_SYSUP,
    "TEXT": KEY_TEXT,
    "TOP": KEY_TOP,
    "UP": KEY_UP,
    "VGA": KEY_VGA,
    "VOLUME_DOWN": KEY_VOLUME_DOWN,
    "VOLUME_UP": KEY_VOLUME_UP,
    "WAKEUP": KEY_WAKEUP,
    "YELLOW": KEY_YELLOW,
}


# Android TV / Fire TV states
STATE_IDLE = "idle"
STATE_OFF = "off"
STATE_PLAYING = "playing"
STATE_PAUSED = "paused"
STATE_STANDBY = "standby"
STATE_STOPPED = "stopped"

#: States that are valid (used by :func:`~androidtv.basetv.state_detection_rules_validator`)
VALID_STATES = (STATE_IDLE, STATE_OFF, STATE_PLAYING, STATE_PAUSED, STATE_STANDBY)

#: Properties that can be used to determine the current state (used by :func:`~androidtv.basetv.state_detection_rules_validator`)
VALID_STATE_PROPERTIES = ("audio_state", "media_session_state")

#: Properties that can be checked for custom state detection (used by :func:`~androidtv.basetv.state_detection_rules_validator`)
VALID_PROPERTIES = VALID_STATE_PROPERTIES + ("wake_lock_size",)

#: The required type for each entry in :py:const:`VALID_PROPERTIES` (used by :func:`~androidtv.basetv.state_detection_rules_validator`)
VALID_PROPERTIES_TYPES = {
    "audio_state": str,
    "media_session_state": int,
    "wake_lock_size": int,
}

# https://developer.android.com/reference/android/media/session/PlaybackState.html
#: States for the :attr:`~androidtv.basetv.basetv.BaseTV.media_session_state` property
MEDIA_SESSION_STATES = {0: None, 1: STATE_STOPPED, 2: STATE_PAUSED, 3: STATE_PLAYING}


# Apps
APP_AE_TV = "com.aetn.aetv.watch"
APP_AMAZON_PRIME_VIDEO = "com.amazon.avod.thirdpartyclient"
APP_AMAZON_VIDEO = "com.amazon.avod"
APP_APPLE_TV_PLUS = "com.apple.atve.android.appletv"
APP_APPLE_TV_PLUS_FIRETV = "com.apple.atve.amazon.appletv"
APP_APPLE_TV_PLUS_SONY = "com.apple.atve.sony.appletv"
APP_ATV_LAUNCHER = "com.google.android.tvlauncher"
APP_BELL_FIBE = "com.quickplay.android.bellmediaplayer"
APP_CBC_GEM = "ca.cbc.android.cbctv"
APP_COMEDY_CENTRAL = "com.vmn.android.comedycentral"
APP_CRAVE = "ca.bellmedia.cravetv"
APP_DAILYMOTION = "com.dailymotion.dailymotion"
APP_DEEZER = "deezer.android.tv"
APP_DISNEY_PLUS = "com.disney.disneyplus"
APP_DISNEY_PLUS_HOTSTAR = "in.startv.hotstar"
APP_DS_PHOTO = "com.synology.dsphoto"
APP_DS_VIDEO = "com.synology.dsvideo"
APP_EMBY = "tv.emby.embyatv"
APP_ES_FILE_EXPLORER = "com.estrongs.android.pop"
APP_FACEBOOK = "com.facebook.katana"
APP_FAWESOME = "com.future.moviesByFawesomeAndroidTV"
APP_FIREFOX = "org.mozilla.tv.firefox"
APP_FIRETV_PACKAGE_LAUNCHER = "com.amazon.tv.launcher"
APP_FIRETV_PACKAGE_SETTINGS = "com.amazon.tv.settings"
APP_FIRETV_STORE = "com.amazon.venezia"
APP_FOOD_NETWORK_GO = "tv.accedo.foodnetwork"
APP_FRANCE_TV = "fr.francetv.pluzz"
APP_GLOBAL_TV = "com.shawmedia.smglobal"
APP_GOOGLE_CAST = "com.google.android.apps.mediashell"
APP_GOOGLE_TV_LAUNCHER = "com.google.android.apps.tv.launcherx"
APP_HAYSTACK_NEWS = "com.haystack.android"
APP_HBO_GO = "eu.hbogo.androidtv.production"
APP_HBO_GO_2 = "com.HBO"
APP_HBO_MAX = "com.hbo.hbonow"
APP_HOICHOI = "com.viewlift.hoichoi"
APP_HULU = "com.hulu.plus"
APP_HULU_2 = "com.hulu.livingroomplus"
APP_HUNGAMA_PLAY = "com.hungama.movies.tv"
APP_IMDB_TV = "com.amazon.imdb.tv.android.app"
APP_IPTV = "ru.iptvremote.android.iptv"
APP_IPTV_SMARTERS_PRO = "com.nst.iptvsmarterstvbox"
APP_JELLYFIN_TV = "org.jellyfin.androidtv"
APP_JIO_CINEMA = "com.jio.media.stb.ondemand"
APP_KODI = "org.xbmc.kodi"
APP_LIVE_CHANNELS = "com.google.android.tv"
APP_MAGENTA_TV = "de.telekom.magentatv.atv"
APP_MIJN_RADIO = "org.samsonsen.nederlandse.radio.holland.nl"
APP_MOLOTOV = "tv.molotov.app"
APP_MRMC = "tv.mrmc.mrmc"
APP_MRMC_LITE = "tv.mrmc.mrmc.lite"
APP_MX_PLAYER = "com.mxtech.videoplayer.ad"
APP_NETFLIX = "com.netflix.ninja"
APP_NLZIET = "nl.nlziet"
APP_NOS = "nl.nos.app"
APP_NPO = "nl.uitzendinggemist"
APP_OCS = "com.orange.ocsgo"
APP_PARAMOUNT_PLUS = "com.cbs.ott"
APP_PEACOCK = "com.peacocktv.peacockandroid"
APP_PLAY_GAMES = "com.google.android.play.games"
APP_PLAY_MUSIC = "com.google.android.music"
APP_PLAY_STORE = "com.android.vending"
APP_PLAY_VIDEOS = "com.google.android.videos"
APP_PLEX = "com.plexapp.android"
APP_PRIME_VIDEO = "com.amazon.amazonvideo.livingroom"
APP_PRIME_VIDEO_FIRETV = "com.amazon.firebat"
APP_SETTINGS = "com.android.tv.settings"
APP_SMART_YOUTUBE_TV = "com.liskovsoft.videomanager"
APP_SONY_ACTION_MENU = "com.sony.dtv.scrums.action"
APP_SONY_ALBUM = "com.sony.dtv.osat.album"
APP_SONY_BRAVIA_SYNC_MENU = "com.sony.dtv.braviasyncmenu"
APP_SONY_BRAVIA_TUTORIALS = "com.sony.dtv.bravialifehack"
APP_SONY_DISCOVER = "com.sony.dtv.discovery"
APP_SONY_HELP = "com.sony.dtv.smarthelp"
APP_SONY_INTERNET_BROWSER = "com.vewd.core.integration.dia"
APP_SONY_LIV = "com.sonyliv"
APP_SONY_MUSIC = "com.sony.dtv.osat.music"
APP_SONY_SCREEN_MIRRORING = "com.sony.dtv.networkapp.wifidirect"
APP_SONY_SELECT = "com.sony.dtv.sonyselect"
APP_SONY_TIMERS = "com.sony.dtv.timers"
APP_SONY_TV = "com.sony.dtv.tvx"
APP_SONY_VIDEO = "com.sony.dtv.osat.video"
APP_SPORT1 = "de.sport1.firetv.video"
APP_SPOTIFY = "com.spotify.tv.android"
APP_STEAM_LINK = "com.valvesoftware.steamlink"
APP_SYFY = "com.amazon.webapps.nbc.syfy"
APP_T2 = "tv.perception.clients.tv.android"
APP_TED = "com.ted.android.tv"
APP_TUNEIN = "tunein.player"
APP_TVHEADEND = "de.cyberdream.dreamepg.tvh.tv.player"
APP_TVNOW = "de.cbc.tvnow.firetv"
APP_TWITCH = "tv.twitch.android.app"
APP_TWITCH_FIRETV = "tv.twitch.android.viewer"
APP_VEVO = "com.vevo.tv"
APP_VH1 = "com.mtvn.vh1android"
APP_VIKI = "com.viki.android"
APP_VIMEO = "com.vimeo.android.videoapp"
APP_VLC = "org.videolan.vlc"
APP_VOYO = "com.phonegap.voyo"
APP_VRV = "com.ellation.vrv"
APP_WAIPU_TV = "de.exaring.waipu.firetv.live"
APP_WATCH_TNT = "com.turner.tnt.android.networkapp"
APP_YOUTUBE = "com.google.android.youtube.tv"
APP_YOUTUBE_FIRETV = "com.amazon.firetv.youtube"
APP_YOUTUBE_KIDS = "com.google.android.youtube.tvkids"
APP_YOUTUBE_KIDS_FIRETV = "com.amazon.firetv.youtube.kids"
APP_YOUTUBE_MUSIC = "com.google.android.youtube.tvmusic"
APP_YOUTUBE_TV = "com.google.android.youtube.tvunplugged"
APP_ZEE5 = "com.graymatrix.did"
APP_ZIGGO_GO_TV = "com.ziggo.tv"
APPS = {
    APP_AE_TV: "A&E",
    APP_AMAZON_PRIME_VIDEO: "Amazon Prime Video",
    APP_AMAZON_VIDEO: "Amazon Video",
    APP_APPLE_TV_PLUS: "Apple TV+",
    APP_APPLE_TV_PLUS_FIRETV: "Apple TV+ (Fire TV)",
    APP_APPLE_TV_PLUS_SONY: "Apple TV+ (Sony)",
    APP_ATV_LAUNCHER: "Android TV Launcher",
    APP_BELL_FIBE: "Bell Fibe",
    APP_CBC_GEM: "CBC Gem",
    APP_COMEDY_CENTRAL: "Comedy Central",
    APP_CRAVE: "Crave",
    APP_DAILYMOTION: "Dailymotion",
    APP_DEEZER: "Deezer",
    APP_DISNEY_PLUS: "Disney+",
    APP_DISNEY_PLUS_HOTSTAR: "Disney+ Hotstar",
    APP_DS_PHOTO: "DS photo",
    APP_DS_VIDEO: "DS video",
    APP_EMBY: "Emby",
    APP_ES_FILE_EXPLORER: "ES File Explorer",
    APP_FACEBOOK: "Facebook Watch",
    APP_FAWESOME: "Fawsome",
    APP_FIREFOX: "Firefox",
    APP_FIRETV_STORE: "FireTV Store",
    APP_FOOD_NETWORK_GO: "Food Network GO",
    APP_FRANCE_TV: "France TV",
    APP_GLOBAL_TV: "Global TV",
    APP_GOOGLE_CAST: "Google Cast",
    APP_GOOGLE_TV_LAUNCHER: "Google TV Launcher",
    APP_HAYSTACK_NEWS: "Haystack News",
    APP_HBO_GO: "HBO GO",
    APP_HBO_GO_2: "HBO GO (2)",
    APP_HBO_MAX: "HBO Max",
    APP_HOICHOI: "Hoichoi",
    APP_HULU: "Hulu",
    APP_HULU_2: "Hulu (2)",
    APP_HUNGAMA_PLAY: "Hungama Play",
    APP_IMDB_TV: "IMDb TV",
    APP_IPTV: "IPTV",
    APP_IPTV_SMARTERS_PRO: "IPTV Smarters Pro",
    APP_JELLYFIN_TV: "Jellyfin",
    APP_JIO_CINEMA: "Jio Cinema",
    APP_KODI: "Kodi",
    APP_LIVE_CHANNELS: "Live Channels",
    APP_MAGENTA_TV: "Magenta TV",
    APP_MIJN_RADIO: "Mijn Radio",
    APP_MOLOTOV: "Molotov",
    APP_MRMC: "MrMC",
    APP_MRMC_LITE: "MrMC Lite",
    APP_MX_PLAYER: "MX Player",
    APP_NETFLIX: "Netflix",
    APP_NLZIET: "NLZIET",
    APP_NOS: "NOS",
    APP_NPO: "NPO",
    APP_OCS: "OCS",
    APP_PARAMOUNT_PLUS: "Paramount+",
    APP_PEACOCK: "Peacock",
    APP_PLAY_GAMES: "Play Games",
    APP_PLAY_MUSIC: "Play Music",
    APP_PLAY_STORE: "Play Store",
    APP_PLAY_VIDEOS: "Play Movies & TV",
    APP_PLEX: "Plex",
    APP_PRIME_VIDEO: "Prime Video",
    APP_PRIME_VIDEO_FIRETV: "Prime Video (FireTV)",
    APP_SETTINGS: "Settings",
    APP_SMART_YOUTUBE_TV: "Smart YouTube TV",
    APP_SONY_ACTION_MENU: "Action Menu",
    APP_SONY_ALBUM: "Album",
    APP_SONY_BRAVIA_SYNC_MENU: "Sync Menu",
    APP_SONY_BRAVIA_TUTORIALS: "BRAVIA Tutorials",
    APP_SONY_DISCOVER: "Discover",
    APP_SONY_HELP: "Help",
    APP_SONY_INTERNET_BROWSER: "Internet Browser",
    APP_SONY_LIV: "SonyLIV",
    APP_SONY_MUSIC: "Music",
    APP_SONY_SCREEN_MIRRORING: "Screen mirroring",
    APP_SONY_SELECT: "Sony Select",
    APP_SONY_TIMERS: "Timers",
    APP_SONY_TV: "TV",
    APP_SONY_VIDEO: "Video",
    APP_SPORT1: "Sport 1",
    APP_SPOTIFY: "Spotify",
    APP_STEAM_LINK: "Steam Link",
    APP_SYFY: "Syfy",
    APP_T2: "T-2 TV",
    APP_TED: "TED",
    APP_TUNEIN: "TuneIn Radio",
    APP_TVHEADEND: "DreamPlayer TVHeadend",
    APP_TVNOW: "RTL Plus",
    APP_TWITCH: "Twitch",
    APP_TWITCH_FIRETV: "Twitch (FireTV)",
    APP_VEVO: "Vevo",
    APP_VH1: "VH1",
    APP_VIKI: "Rakuten Viki",
    APP_VIMEO: "Vimeo",
    APP_VLC: "VLC",
    APP_VOYO: "VOYO",
    APP_VRV: "VRV",
    APP_WAIPU_TV: "Waipu TV",
    APP_WATCH_TNT: "Watch TNT",
    APP_YOUTUBE: "YouTube",
    APP_YOUTUBE_FIRETV: "YouTube (FireTV)",
    APP_YOUTUBE_KIDS: "YouTube Kids",
    APP_YOUTUBE_KIDS_FIRETV: "YouTube Kids (FireTV)",
    APP_YOUTUBE_MUSIC: "YouTube Music",
    APP_YOUTUBE_TV: "YouTube TV",
    APP_ZEE5: "ZEE5",
    APP_ZIGGO_GO_TV: "Ziggo GO TV",
}


# Regular expressions
REGEX_MEDIA_SESSION_STATE = re.compile(r"state=(?P<state>[0-9]+)", re.MULTILINE)
REGEX_WAKE_LOCK_SIZE = re.compile(r"size=(?P<size>[0-9]+)")

# Regular expression patterns
DEVICE_REGEX_PATTERN = r"Devices: (.*?)\W"
MAC_REGEX_PATTERN = "ether (.*?) brd"
MAX_VOLUME_REGEX_PATTERN = r"Max: (\d{1,})"
MUTED_REGEX_PATTERN = r"Muted: (.*?)\W"
STREAM_MUSIC_REGEX_PATTERN = "STREAM_MUSIC(.*?)- STREAM"
VOLUME_REGEX_PATTERN = r"\): (\d{1,})"

#: Default authentication timeout (in s) for :meth:`adb_shell.handle.tcp_handle.TcpHandle.connect` and :meth:`adb_shell.handle.tcp_handle_async.TcpHandleAsync.connect`
DEFAULT_AUTH_TIMEOUT_S = 10.0

#: Default transport timeout (in s) for :meth:`adb_shell.handle.tcp_handle.TcpHandle.connect` and :meth:`adb_shell.handle.tcp_handle_async.TcpHandleAsync.connect`
DEFAULT_TRANSPORT_TIMEOUT_S = 1.0

#: Default timeout (in s) for :class:`adb_shell.handle.tcp_handle.TcpHandle` and :class:`adb_shell.handle.tcp_handle_async.TcpHandleAsync`
DEFAULT_ADB_TIMEOUT_S = 9.0

#: Default timeout for acquiring the lock that protects ADB commands
DEFAULT_LOCK_TIMEOUT_S = 3.0



import sys
import unittest

try:
    # Python3
    from unittest.mock import patch
except ImportError:
    # Python2
    from mock import patch


sys.path.insert(0, "..")

from androidtv import setup
from androidtv.androidtv.androidtv_sync import AndroidTVSync
from androidtv.firetv.firetv_sync import FireTVSync
from . import patchers


DEVICE_PROPERTIES_OUTPUT1 = "Amazon\n\n\n123"

DEVICE_PROPERTIES_DICT1 = {
    "manufacturer": "Amazon",
    "model": "",
    "serialno": None,
    "sw_version": "123",
    "wifimac": None,
    "ethmac": None,
}

DEVICE_PROPERTIES_OUTPUT2 = "Not Amazon\n\n\n456"

DEVICE_PROPERTIES_DICT2 = {
    "manufacturer": "Not Amazon",
    "model": "",
    "serialno": None,
    "sw_version": "456",
    "wifimac": None,
    "ethmac": None,
}


class TestSetup(unittest.TestCase):
    PATCH_KEY = "python"

    def test_setup(self):
        """Test that the ``setup`` function works correctly."""
        with self.assertRaises(ValueError):
            setup("HOST", 5555, device_class="INVALID")

        with patchers.PATCH_ADB_DEVICE_TCP, patchers.patch_connect(True)[self.PATCH_KEY], patchers.patch_shell(
            DEVICE_PROPERTIES_OUTPUT1
        )[self.PATCH_KEY]:
            ftv = setup("HOST", 5555)
            self.assertIsInstance(ftv, FireTVSync)
            self.assertDictEqual(ftv.device_properties, DEVICE_PROPERTIES_DICT1)

        with patchers.PATCH_ADB_DEVICE_TCP, patchers.patch_connect(True)[self.PATCH_KEY], patchers.patch_shell(
            DEVICE_PROPERTIES_OUTPUT2
        )[self.PATCH_KEY]:
            atv = setup("HOST", 5555)
            self.assertIsInstance(atv, AndroidTVSync)
            self.assertDictEqual(atv.device_properties, DEVICE_PROPERTIES_DICT2)

        with patchers.PATCH_ADB_DEVICE_TCP, patchers.patch_connect(True)[self.PATCH_KEY], patchers.patch_shell(
            DEVICE_PROPERTIES_OUTPUT1
        )[self.PATCH_KEY]:
            ftv = setup("HOST", 5555, device_class="androidtv")
            self.assertIsInstance(ftv, AndroidTVSync)
            self.assertDictEqual(ftv.device_properties, DEVICE_PROPERTIES_DICT1)

        with patchers.PATCH_ADB_DEVICE_TCP, patchers.patch_connect(True)[self.PATCH_KEY], patchers.patch_shell(
            DEVICE_PROPERTIES_OUTPUT2
        )[self.PATCH_KEY]:
            atv = setup("HOST", 5555, device_class="firetv")
            self.assertIsInstance(atv, FireTVSync)
            self.assertDictEqual(atv.device_properties, DEVICE_PROPERTIES_DICT2)


if __name__ == "__main__":
    unittest.main()



import asyncio
from contextlib import asynccontextmanager
import sys
import unittest
from unittest.mock import patch

sys.path.insert(0, "..")

from androidtv.adb_manager.adb_manager_async import _acquire, ADBPythonAsync, ADBServerAsync
from androidtv.exceptions import LockNotAcquiredException

from . import async_patchers
from .async_wrapper import awaiter


class Read(object):
    """Mock an opened file that can be read."""

    async def read(self):
        return ""


class ReadFail(object):
    """Mock an opened file that cannot be read."""

    async def read(self):
        raise FileNotFoundError


PNG_IMAGE = b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\n\x00\x00\x00\n\x08\x06\x00\x00\x00\x8d2\xcf\xbd\x00\x00\x00\x04sBIT\x08\x08\x08\x08|\x08d\x88\x00\x00\x00\tpHYs\x00\x00\x0fa\x00\x00\x0fa\x01\xa8?\xa7i\x00\x00\x00\x0eIDAT\x18\x95c`\x18\x05\x83\x13\x00\x00\x01\x9a\x00\x01\x16\xca\xd3i\x00\x00\x00\x00IEND\xaeB`\x82"

PNG_IMAGE_NEEDS_REPLACING = PNG_IMAGE[:5] + b"\r" + PNG_IMAGE[5:]


@asynccontextmanager
async def open_priv(infile):
    """A patch that will read the private key but not the public key."""
    try:
        if infile == "adbkey":
            yield Read()
        else:
            yield ReadFail()
    finally:
        pass


@asynccontextmanager
async def open_priv_pub(infile):
    try:
        yield Read()
    finally:
        pass


class AsyncFakeLock:
    def __init__(self):
        self._acquired = True

    async def acquire(self):
        if self._acquired:
            self._acquired = False
            return True
        return self._acquired

    def release(self):
        self._acquired = True


class AsyncLockedLock(AsyncFakeLock):
    def __init__(self):
        self._acquired = False


class AsyncTimedLock(AsyncFakeLock):
    async def acquire(self):
        await asyncio.sleep(1.0)
        return await super().acquire()


class TestLock(unittest.TestCase):
    """Test the async lock code."""

    @awaiter
    async def test_succeed(self):
        lock = AsyncFakeLock()
        async with _acquire(lock):
            self.assertTrue(True)
            return
        self.assertTrue(False)

    @awaiter
    async def test_fail(self):
        lock = AsyncLockedLock()
        with self.assertRaises(LockNotAcquiredException):
            async with _acquire(lock):
                pass  # self.assertTrue(False)

    @awaiter
    async def test_fail_timeout(self):
        lock = AsyncTimedLock()
        with self.assertRaises(LockNotAcquiredException):
            async with _acquire(lock, 0.1):
                pass  # self.assertTrue(False)


class TestADBPythonAsync(unittest.TestCase):
    """Test the `ADBPythonAsync` class."""

    PATCH_KEY = "python"

    def setUp(self):
        """Create an `ADBPythonAsync` instance."""
        with async_patchers.PATCH_ADB_DEVICE_TCP, async_patchers.patch_connect(True)[self.PATCH_KEY]:
            self.adb = ADBPythonAsync("HOST", 5555)

    @awaiter
    async def test_connect_success(self):
        """Test when the connect attempt is successful."""
        with async_patchers.patch_connect(True)[self.PATCH_KEY]:
            self.assertTrue(await self.adb.connect())
            self.assertTrue(self.adb.available)

    @awaiter
    async def test_connect_fail(self):
        """Test when the connect attempt fails."""
        with async_patchers.patch_connect(False)[self.PATCH_KEY]:
            self.assertFalse(await self.adb.connect())
            self.assertFalse(self.adb.available)

        with async_patchers.patch_connect(True)[self.PATCH_KEY]:
            self.assertTrue(await self.adb.connect())
            self.assertTrue(self.adb.available)

        with async_patchers.patch_connect(True)[self.PATCH_KEY]:
            with async_patchers.PATCH_CONNECT_FAIL_CUSTOM_EXCEPTION[self.PATCH_KEY]:
                self.assertFalse(await self.adb.connect())
                self.assertFalse(self.adb.available)

    @awaiter
    async def test_connect_fail_lock(self):
        """Test when the connect attempt fails due to the lock."""
        with async_patchers.patch_connect(True)[self.PATCH_KEY]:
            with patch.object(self.adb, "_adb_lock", AsyncLockedLock()):
                self.assertFalse(await self.adb.connect())
                self.assertFalse(self.adb.available)

    @awaiter
    async def test_adb_shell_fail(self):
        """Test when an ADB shell command is not sent because the device is unavailable."""
        self.assertFalse(self.adb.available)
        with async_patchers.patch_connect(True)[self.PATCH_KEY], async_patchers.patch_shell(None)[self.PATCH_KEY]:
            self.assertIsNone(await self.adb.shell("TEST"))

        with async_patchers.patch_connect(True)[self.PATCH_KEY], async_patchers.patch_shell("TEST")[self.PATCH_KEY]:
            self.assertTrue(await self.adb.connect())
            with patch.object(self.adb, "_adb_lock", AsyncLockedLock()):
                with self.assertRaises(LockNotAcquiredException):
                    await self.adb.shell("TEST")

                with self.assertRaises(LockNotAcquiredException):
                    await self.adb.shell("TEST2")

    @awaiter
    async def test_adb_shell_success(self):
        """Test when an ADB shell command is successfully sent."""
        with async_patchers.patch_connect(True)[self.PATCH_KEY], async_patchers.patch_shell("TEST")[self.PATCH_KEY]:
            self.assertTrue(await self.adb.connect())
            self.assertEqual(await self.adb.shell("TEST"), "TEST")

    @awaiter
    async def test_adb_shell_fail_lock_released(self):
        """Test that the ADB lock gets released when an exception is raised."""
        with async_patchers.patch_connect(True)[self.PATCH_KEY], async_patchers.patch_shell("TEST")[self.PATCH_KEY]:
            self.assertTrue(await self.adb.connect())

        with async_patchers.patch_shell("TEST", error=True)[self.PATCH_KEY], patch.object(
            self.adb, "_adb_lock", AsyncFakeLock()
        ):
            with patch("{}.AsyncFakeLock.release".format(__name__)) as release:
                with self.assertRaises(Exception):
                    await self.adb.shell("TEST")
                assert release.called

    @awaiter
    async def test_adb_shell_lock_not_acquired_not_released(self):
        """Test that the lock does not get released if it is not acquired."""
        with async_patchers.patch_connect(True)[self.PATCH_KEY], async_patchers.patch_shell("TEST")[self.PATCH_KEY]:
            self.assertTrue(await self.adb.connect())
            self.assertEqual(await self.adb.shell("TEST"), "TEST")

        with async_patchers.patch_shell("TEST")[self.PATCH_KEY], patch.object(self.adb, "_adb_lock", AsyncLockedLock()):
            with patch("{}.AsyncLockedLock.release".format(__name__)) as release:
                with self.assertRaises(LockNotAcquiredException):
                    await self.adb.shell("TEST")

                release.assert_not_called()

    @awaiter
    async def test_adb_push_fail(self):
        """Test when an ADB push command is not executed because the device is unavailable."""
        self.assertFalse(self.adb.available)
        with async_patchers.patch_connect(True)[self.PATCH_KEY]:
            with async_patchers.PATCH_PUSH[self.PATCH_KEY] as patch_push:
                await self.adb.push("TEST_LOCAL_PATCH", "TEST_DEVICE_PATH")
                patch_push.assert_not_called()

        with async_patchers.patch_connect(True)[self.PATCH_KEY]:
            with async_patchers.PATCH_PUSH[self.PATCH_KEY] as patch_push:
                self.assertTrue(await self.adb.connect())
                with patch.object(self.adb, "_adb_lock", AsyncLockedLock()):
                    with self.assertRaises(LockNotAcquiredException):
                        await self.adb.push("TEST_LOCAL_PATH", "TEST_DEVICE_PATH")

                    patch_push.assert_not_called()

    @awaiter
    async def test_adb_push_success(self):
        """Test when an ADB push command is successfully executed."""
        with async_patchers.patch_connect(True)[self.PATCH_KEY]:
            with async_patchers.PATCH_PUSH[self.PATCH_KEY] as patch_push:
                self.assertTrue(await self.adb.connect())
                await self.adb.push("TEST_LOCAL_PATH", "TEST_DEVICE_PATH")
                self.assertEqual(patch_push.call_count, 1)

    @awaiter
    async def test_adb_pull_fail(self):
        """Test when an ADB pull command is not executed because the device is unavailable."""
        self.assertFalse(self.adb.available)
        with async_patchers.patch_connect(True)[self.PATCH_KEY]:
            with async_patchers.PATCH_PULL[self.PATCH_KEY] as patch_pull:
                await self.adb.pull("TEST_LOCAL_PATCH", "TEST_DEVICE_PATH")
                patch_pull.assert_not_called()

        with async_patchers.patch_connect(True)[self.PATCH_KEY]:
            with async_patchers.PATCH_PULL[self.PATCH_KEY] as patch_pull:
                self.assertTrue(await self.adb.connect())
                with patch.object(self.adb, "_adb_lock", AsyncLockedLock()):
                    with self.assertRaises(LockNotAcquiredException):
                        await self.adb.pull("TEST_LOCAL_PATH", "TEST_DEVICE_PATH")
                    patch_pull.assert_not_called()

    @awaiter
    async def test_adb_pull_success(self):
        """Test when an ADB pull command is successfully executed."""
        with async_patchers.patch_connect(True)[self.PATCH_KEY]:
            with async_patchers.PATCH_PULL[self.PATCH_KEY] as patch_pull:
                self.assertTrue(await self.adb.connect())
                await self.adb.pull("TEST_LOCAL_PATH", "TEST_DEVICE_PATH")
                self.assertEqual(patch_pull.call_count, 1)

    @awaiter
    async def test_adb_screencap_fail_unavailable(self):
        """Test when an ADB screencap command fails because the connection is unavailable."""
        self.assertFalse(self.adb.available)
        self.assertIsNone(await self.adb.screencap())

    @awaiter
    async def test_adb_screencap_lock_not_acquired(self):
        """Test when an ADB screencap command fails because the ADB lock could not be acquired."""
        with async_patchers.patch_connect(True)[self.PATCH_KEY], async_patchers.patch_shell("TEST")[self.PATCH_KEY]:
            self.assertTrue(await self.adb.connect())
            self.assertEqual(await self.adb.shell("TEST"), "TEST")

        with async_patchers.patch_shell(PNG_IMAGE)[self.PATCH_KEY], patch.object(
            self.adb, "_adb_lock", AsyncLockedLock()
        ):
            with patch("{}.AsyncLockedLock.release".format(__name__)) as release:
                with self.assertRaises(LockNotAcquiredException):
                    await self.adb.screencap()

                release.assert_not_called()

    @awaiter
    async def test_adb_screencap_success(self):
        """Test the `screencap` method."""
        with async_patchers.patch_connect(True)[self.PATCH_KEY], async_patchers.patch_shell(PNG_IMAGE)[self.PATCH_KEY]:
            self.assertTrue(await self.adb.connect())

            if isinstance(self.adb, ADBPythonAsync):
                self.assertEqual(await self.adb.screencap(), PNG_IMAGE)

                with async_patchers.patch_shell(PNG_IMAGE_NEEDS_REPLACING)[self.PATCH_KEY]:
                    self.assertEqual(await self.adb.screencap(), PNG_IMAGE)

            else:
                with patch.object(
                    self.adb._adb_device, "screencap", return_value=PNG_IMAGE, new_callable=async_patchers.AsyncMock
                ):
                    self.assertEqual(await self.adb.screencap(), PNG_IMAGE)


class TestADBPythonUsbAsync(unittest.TestCase):
    """Test the `ADBPythonAsync` class using a USB connection."""

    def test_init(self):
        """Create an `ADBPythonSync` instance with a USB connection."""
        with patch("androidtv.adb_manager.adb_manager_async.AdbDeviceUsbAsync") as patched:
            ADBPythonAsync("", 5555)
            assert patched.called


class TestADBServerAsync(TestADBPythonAsync):
    """Test the `ADBServerAsync` class."""

    PATCH_KEY = "server"

    def setUp(self):
        """Create an `ADBServerAsync` instance."""
        self.adb = ADBServerAsync("HOST", 5555, "ADB_SERVER_IP")

    @awaiter
    async def test_connect_fail_server(self):
        """Test that the ``connect`` method works correctly."""
        with async_patchers.patch_connect(True)[self.PATCH_KEY]:
            self.assertTrue(await self.adb.connect())

            with async_patchers.PATCH_ADB_SERVER_RUNTIME_ERROR:
                self.assertFalse(await self.adb.connect())
                self.assertFalse(self.adb.available)


class TestADBPythonAsyncWithAuthentication(unittest.TestCase):
    """Test the `ADBPythonAsync` class."""

    PATCH_KEY = "python"

    def setUp(self):
        """Create an `ADBPythonAsync` instance."""
        with async_patchers.PATCH_ADB_DEVICE_TCP, async_patchers.patch_connect(True)[self.PATCH_KEY]:
            self.adb = ADBPythonAsync("HOST", 5555, "adbkey")

    @awaiter
    async def test_connect_success_with_priv_key(self):
        """Test when the connect attempt is successful when using a private key."""
        with async_patchers.patch_connect(True)[self.PATCH_KEY], patch(
            "androidtv.adb_manager.adb_manager_async.aiofiles.open", open_priv
        ), patch("androidtv.adb_manager.adb_manager_async.PythonRSASigner", return_value="TEST"):
            self.assertTrue(await self.adb.connect())
            self.assertTrue(self.adb.available)

        with async_patchers.patch_connect(True)[self.PATCH_KEY]:
            with patch("androidtv.adb_manager.adb_manager_async.aiofiles.open") as patch_open:
                self.assertTrue(await self.adb.connect())
                self.assertTrue(self.adb.available)
                assert not patch_open.called

    @awaiter
    async def test_connect_success_with_priv_pub_key(self):
        """Test when the connect attempt is successful when using private and public keys."""
        with async_patchers.patch_connect(True)[self.PATCH_KEY], patch(
            "androidtv.adb_manager.adb_manager_async.aiofiles.open", open_priv_pub
        ), patch("androidtv.adb_manager.adb_manager_async.PythonRSASigner", return_value=None):
            self.assertTrue(await self.adb.connect())
            self.assertTrue(self.adb.available)


class TestADBPythonAsyncClose(unittest.TestCase):
    """Test the `ADBPythonAsync.close` method."""

    PATCH_KEY = "python"

    @awaiter
    async def test_close(self):
        """Test the `ADBPythonAsync.close` method."""
        with async_patchers.PATCH_ADB_DEVICE_TCP, async_patchers.patch_connect(True)[self.PATCH_KEY]:
            self.adb = ADBPythonAsync("HOST", 5555)

        with async_patchers.patch_connect(True)[self.PATCH_KEY]:
            self.assertTrue(await self.adb.connect())
            self.assertTrue(self.adb.available)

            await self.adb.close()
            self.assertFalse(self.adb.available)
